# -*- coding: utf-8 -*-

# @Time    : 2019/11/12 15:37
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause
import functools
import operator
import random
import sys
import warnings
from copy import deepcopy

import numpy as np
import pandas as pd
import sympy
from deap.tools import Logbook
from scipy import optimize
from sklearn.exceptions import DataConversionWarning
from sklearn.metrics import r2_score
from sklearn.utils import check_array
from sympy import sympify

from featurebox.combination.dim import dim_func, dnan, Dim
from featurebox.tools.exports import Store


def calcualte_dim(expr01, rep_list, dim_list):
    func0 = sympy.utilities.lambdify(rep_list, expr01, modules=[dim_func(), 'numpy'])
    dim_ = func0(*dim_list)
    return dim_


def _compile(expr_, pset):
    """

    Parameters
    ----------
    expr_
    pset

    Returns
    -------

    """
    code = str(expr_)
    if len(pset.arguments) > 0:
        # This section is a stripped version of the lambdify
        # function of SymPy 0.6.6.
        args = ",".join(arg for arg in pset.arguments)
        code = "lambda {args}: {code}".format(args=args, code=code)
    try:
        return eval(code, pset.context, {})
    except MemoryError:
        _, _, traceback = sys.exc_info()
        raise MemoryError("DEAP : Error in tree evaluation :"
                          " Python cannot evaluate a tree higher than 90. "
                          "To avoid this problem, you should use bloat control on your "
                          "operators. See the DEAP documentation for more information. "
                          "DEAP will now abort.").with_traceback(traceback)


def addCoefficient(expr01, inter_add=True, iner_add=True, random_add=None):
    """

    Parameters
    ----------
    expr01
    inter_add
    iner_add
    random_add

    Returns
    -------

    """

    def get_args(expr_):
        """"""
        list_arg = []
        for i in expr_.args:
            list_arg.append(i)
            if i.args:
                re = get_args(i)
                list_arg.extend(re)

        return list_arg

    arg_list = get_args(expr01)
    arg_list = [i for i in arg_list if i not in expr01.args]
    cho = []
    a_list = []
    #

    if isinstance(expr01, sympy.Add):

        for i, j in enumerate(expr01.args):
            Wi = sympy.Symbol("W%s" % i)
            expr01 = expr01.subs(j, Wi * j)
            a_list.append(Wi)

    else:

        A = sympy.Symbol("A")
        expr01 = sympy.Mul(expr01, A)

        a_list.append(A)

    if inter_add:
        B = sympy.Symbol("B")
        expr01 = expr01 + B
        a_list.append(B)

    if iner_add:
        cho_add = [i.args for i in arg_list if isinstance(i, sympy.Add)]
        cho_add = [[_ for _ in cho_addi if not _.is_number] for cho_addi in cho_add]
        [cho.extend(i) for i in cho_add]

    if random_add:
        pass
    #     lest = [i for i in arg_list if i not in cho]
    #     if len(lest) != 0:
    #         cho2 = random.sample(lest, 1)
    #         cho.extend(cho2)

    a_cho = [sympy.Symbol("k%s" % i) for i in range(len(cho))]
    for ai, choi in zip(a_cho, cho):
        expr01 = expr01.subs(choi, ai * choi)
    a_list.extend(a_cho)

    return expr01, a_list


def calculateExpr(expr01, x, y, terminals, scoring=None, add_coeff=True,
                  del_no_important=False, filter_warning=True, inter_add=True, iner_add=True, random_add=None):
    """

    Parameters
    ----------
    random_add
    iner_add
    inter_add
    expr01
    x
    y
    scoring
    add_coeff
    del_no_important
    filter_warning
    terminals

    Returns
    -------

    """

    def split_x(x):
        if x.ndim == 1:
            return [x]
        else:
            return [*x.T]

    if filter_warning:
        warnings.filterwarnings("ignore")
    if not scoring:
        scoring = r2_score

    expr00 = deepcopy(expr01)  #

    if add_coeff:

        expr01, a_list = addCoefficient(expr01, inter_add=inter_add, iner_add=iner_add, random_add=random_add)

        try:
            func0 = sympy.utilities.lambdify(terminals + a_list, expr01)

            def func(x_, p):
                """"""
                num_list = []

                num_list.extend(split_x(x))

                num_list.extend(p)
                return func0(*num_list)

            def res(p, x_, y_):
                """"""
                return y_ - func(x_, p)

            result = optimize.least_squares(res, x0=[1] * len(a_list), args=(x, y), loss='linear', ftol=1e-3)

            cof = result.x
            cof_ = []
            for a_listi, cofi in zip(a_list, cof):
                if "A" or "W" in a_listi.name:
                    cof_.append(cofi)
                else:
                    cof_.append(np.round(cofi, decimals=3))
            cof = cof_
            for ai, choi in zip(a_list, cof):
                expr01 = expr01.subs(ai, choi)
        except (ValueError, KeyError, NameError, TypeError):
            expr01 = expr00

    else:
        pass  #

    try:
        if del_no_important and isinstance(expr01, sympy.Add) and len(expr01.args) >= 3:
            re_list = []
            for expri in expr01.args:
                if not isinstance(expri, sympy.Float):
                    func0 = sympy.utilities.lambdify(terminals, expri)
                    re = np.mean(func0(*split_x(x)))
                    if abs(re) > abs(0.001 * np.mean(y)):
                        re_list.append(expri)
                else:
                    re_list.append(expri)
            expr01 = sum(re_list)
        else:
            pass

        func0 = sympy.utilities.lambdify(terminals, expr01)
        try:
            re = func0(*split_x(x))
        except:
            print(expr01)
        sp = split_x(x)
        re = func0(*sp)
        re = re.ravel()
        assert y.shape == re.shape
        # assert_all_finite(re)
        re = check_array(re, ensure_2d=False)
        score = scoring(y, re)

    except (ValueError, DataConversionWarning, NameError, KeyError, AssertionError, AttributeError):
        score = -0
    else:
        if np.isnan(score):
            score = -0
    return score, expr01


def calculatePrecision(individual, pset, x, y, scoring=None, add_coeff=True, filter_warning=True, cal_dim=True,
                       inter_add=True, iner_add=True, random_add=None):
    """

    Parameters
    ----------
    inter_add
    iner_add
    random_add
    cal_dim
    scoring
    individual
    pset
    x
    y
    add_coeff
    filter_warning

    Returns
    -------

    """

    if scoring is None:
        scoring = r2_score
    # '''1 not expand'''
    expr_no = sympify(_compile(individual, pset))
    # '''2 expand by sympy.expand,long expr is slow, use if when needed'''
    # expr_no = sympy.expand(compile_(individual, pset), deep=False, power_base=False, power_exp=False, mul=True,
    #                        log=False, multinomial=False)

    terminals = pset.rep_name_list

    score, expr = calculateExpr(expr_no, x, y, terminals=terminals, scoring=scoring, add_coeff=add_coeff,
                                filter_warning=filter_warning, inter_add=inter_add, iner_add=iner_add,
                                random_add=random_add)

    if cal_dim:
        rep_list = pset.rep_name_list
        dim_list = pset.dim_list
        try:
            dim = calcualte_dim(expr, rep_list, dim_list)
        except TypeError:
            dim = dnan
    else:
        dim = dnan

    if isinstance(dim, (float, int)) or dim is None:
        dim = dnan
        withdim = 0

    elif isinstance(dim, Dim) and not dim.anyisnan:
        withdim = 1
    else:
        dim = dnan
        withdim = 0

    return score, expr, dim, withdim


def custom_loss_func(y_true, y_pred):
    """"""
    diff = - np.abs(y_true - y_pred) / y_true + 1
    return np.mean(diff)


def getName(x):
    """"""
    if isinstance(x, pd.DataFrame):
        name = x.columns.values
        name = [sympy.Symbol(i) for i in name]
        rep_name = [sympy.Symbol("x%d" % i) for i in range(len(name))]

    elif isinstance(x, np.ndarray):
        check_array(x)
        name = x.shape[1]
        name = [sympy.Symbol("x%d" % i) for i in range(name)]
        rep_name = [sympy.Symbol("x%d" % i) for i in range(len(name))]
    else:
        raise TypeError("just support np.ndarray and pd.DataFrame")

    return name, rep_name


def sub(expr01, subed, subs):
    """"""
    listt = list(zip(subed, subs))
    return expr01.subs(listt)


def varAnd(population, toolbox, cxpb, mutpb):
    rst = random.getstate()
    offspring = [toolbox.clone(ind) for ind in population]
    random.setstate(rst)
    # Apply crossover and mutation on the offspring
    for i in range(1, len(offspring), 2):
        if random.random() < cxpb:
            offspring[i - 1], offspring[i] = toolbox.mate(offspring[i - 1],
                                                          offspring[i])
            del offspring[i - 1].fitness.values, offspring[i].fitness.values
    for i in range(len(offspring)):
        if random.random() < mutpb:
            offspring[i], = toolbox.mutate(offspring[i])
            del offspring[i].fitness.values
    return offspring


# def eaSimple(population, toolbox, cxpb, mutpb, ngen, stats=None,
#              halloffame=None, verbose=__debug__, pset=None, store=True):
#     """
#
#     Parameters
#     ----------
#     population
#     toolbox
#     cxpb
#     mutpb
#     ngen
#     stats
#     halloffame
#     verbose
#     pset
#     store
#     Returns
#     -------
#
#     """
#     rst = random.getstate()
#     len_pop = len(population)
#     logbook = Logbook()
#     logbook.header = ['gen', 'pop'] + (stats.fields if stats else [])
#
#     # Evaluate the individuals with an invalid fitness
#     invalid_ind = [ind for ind in population if not ind.fitness.valid]
#
#     # fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
#     fitnesses = toolbox.parallel(iterable=population)
#     for ind, fit, in zip(invalid_ind, fitnesses):
#         ind.fitness.values = fit[0],
#         ind.expr = fit[1]
#         ind.dim = fit[2]
#         ind.withdim = fit[3]
#
#     add_ind = toolbox.select_kbest_target_dim(population, K_best=0.1 * len_pop)
#     if halloffame is not None:
#         halloffame.update(add_ind)
#
#     record = stats.compile(population) if stats else {}
#     logbook.record(gen=0, nevals=len(population), **record)
#     if verbose:
#         print(logbook.stream)
#     data_all = {}
#
#     # Begin the generational process
#     random.setstate(rst)
#     for gen in range(1, ngen + 1):
#         rst = random.getstate()
#
#         if store:
#             rst = random.getstate()
#             target_dim = toolbox.select_kbest_target_dim.keywords['dim_type']
#             subp = functools.partial(sub, subed=pset.rep_name_list, subs=pset.real_name_list)
#             datamnist = {"gen{}_pop{}".format(gen, n): {"gen": gen, "pop": n,
#                                                    "score": i.fitness.values[0],
#                                                    "expr": str(subp(i.expr)),
#                                                    "with_dim": 1 if i.withdim else 0,
#                                                    "dim_is_target_dim": 1 if i.dim in target_dim else 0,
#                                                    "gen_dim": "{}{}".format(gen, 1 if i.withdim else 0),
#                                                    "gen_target_dim": "{}{}".format(gen,
#                                                                                    1 if i.dim in target_dim else 0),
#                                                    "socre_dim": i.fitness.values[0] if i.withdim else 0,
#                                                    "socre_target_dim": i.fitness.values[
#                                                        0] if i.dim in target_dim else 0,
#                                                    } for n, i in enumerate(population) if i is not None}
#             data_all.update(datamnist)
#         random.setstate(rst)
#         # select_gs the next generation individuals
#         offspring = toolbox.select_gs(population, len_pop)
#
#         # Vary the pool of individuals
#         offspring = varAnd(offspring, toolbox, cxpb, mutpb)
#
#         rst = random.getstate()
#
#         # Evaluate the individuals with an invalid fitness
#         invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
#         # fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
#         # fitnesses = parallelize(n_jobs=3, func=toolbox.evaluate, iterable=invalid_ind,  respective=False)
#         fitnesses = toolbox.parallel(iterable=invalid_ind)
#         for ind, fit in zip(invalid_ind, fitnesses):
#             ind.fitness.values = fit[0],
#             ind.expr = fit[1]
#             ind.dim = fit[2]
#             ind.withdim = fit[3]
#
#         add_ind = toolbox.select_kbest_target_dim(population, K_best=0.1 * len_pop)
#         # add_ind2 = toolbox.select_kbest_dimless(population, K_best=0.2 * len_pop)
#         # add_ind3 = toolbox.select_kbest(population, K_best=5)
#         offspring += add_ind
#         # offspring += add_ind2
#         # offspring += add_ind3
#
#         # Update the hall of fame with the generated individuals
#         if halloffame is not None:
#             halloffame.update(add_ind)
#             if len(halloffame.items) > 0 and halloffame.items[-1].fitness.values[0] >= 0.95:
#                 print(halloffame.items[-1])
#                 print(halloffame.items[-1].fitness.values[0])
#                 break
#
#         population[:] = offspring
#         # Append the current generation statistics to the logbook
#         record = stats.compile(population) if stats else {}
#         logbook.record(gen=gen, nevals=len(population), **record)
#         if verbose:
#             print(logbook.stream)
#         random.setstate(rst)
#
#     store = Store()
#     store.to_csv(data_all)
#     return population, logbook

def eaSimple(population, toolbox, cxpb, mutpb, ngen, stats=None,
             halloffame=None, verbose=__debug__, pset=None, store=True):
    """

    Parameters
    ----------
    population
    toolbox
    cxpb
    mutpb
    ngen
    stats
    halloffame
    verbose
    pset
    store
    Returns
    -------

    """
    rst = random.getstate()
    len_pop = len(population)
    logbook = Logbook()
    logbook.header = [] + (stats.fields if stats else [])
    data_all = {}
    random.setstate(rst)

    for gen in range(1, ngen + 1):
        "评价"
        rst = random.getstate()
        """score"""
        invalid_ind = [ind for ind in population if not ind.fitness.valid]
        fitnesses = toolbox.parallel(iterable=population)
        for ind, fit, in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit[0],
            ind.expr = fit[1]
            ind.y_dim = fit[2]
            ind.withdim = fit[3]
        random.setstate(rst)

        rst = random.getstate()
        """elite"""
        add_ind = []
        add_ind1 = toolbox.select_kbest_target_dim(population, K_best=0.1 * len_pop)
        add_ind2 = toolbox.select_kbest_dimless(population, K_best=0.1 * len_pop)
        add_ind3 = toolbox.select_kbest(population, K_best=5)
        add_ind += add_ind1
        add_ind += add_ind2
        add_ind += add_ind3
        elite_size = len(add_ind)
        random.setstate(rst)

        rst = random.getstate()
        """score"""
        if store:
            target_dim = toolbox.select_kbest_target_dim.keywords['dim_type']
            subp = functools.partial(sub, subed=pset.rep_name_list, subs=pset.real_name_list)
            data = {"gen{}_pop{}".format(gen, n): {"gen": gen, "pop": n,
                                                   "score": i.fitness.values[0],
                                                   "expr": str(subp(i.expr)),
                                                   "with_dim": 1 if i.withdim else 0,
                                                   "dim_is_target_dim": 1 if i.y_dim in target_dim else 0,
                                                   "gen_dim": "{}{}".format(gen, 1 if i.withdim else 0),
                                                   "gen_target_dim": "{}{}".format(gen,
                                                                                   1 if i.y_dim in target_dim else 0),
                                                   "socre_dim": i.fitness.values[0] if i.withdim else 0,
                                                   "socre_target_dim": i.fitness.values[
                                                       0] if i.y_dim in target_dim else 0,
                                                   } for n, i in enumerate(population) if i is not None}
            data_all.update(data)
        random.setstate(rst)

        rst = random.getstate()
        """record"""
        if halloffame is not None:
            halloffame.update(add_ind1)
            if len(halloffame.items) > 0 and halloffame.items[-1].fitness.values[0] >= 0.95:
                print(halloffame.items[-1])
                print(halloffame.items[-1].fitness.values[0])
                break
        random.setstate(rst)

        rst = random.getstate()
        """Dynamic output"""

        record = stats.compile_(population) if stats else {}
        logbook.record(gen=gen, pop=len(population), **record)

        if verbose:
            print(logbook.stream)
        random.setstate(rst)

        """crossover, mutate"""
        offspring = toolbox.select_gs(population, len_pop - elite_size)
        # Vary the pool of individuals
        offspring = varAnd(offspring, toolbox, cxpb, mutpb)

        rst = random.getstate()
        """re-run"""
        offspring.extend(add_ind)
        population[:] = offspring
        random.setstate(rst)

    store = Store()
    store.to_csv(data_all)
    return population, logbook

    #
    # # Begin the generational process
    # random.setstate(rst)
    # for gen in range(1, ngen + 1):
    #     rst = random.getstate()
    #
    #     if store:
    #         rst = random.getstate()
    #         target_dim = toolbox.select_kbest_target_dim.keywords['dim_type']
    #         subp = functools.partial(sub, subed=pset.rep_name_list, subs=pset.real_name_list)
    #         datamnist = {"gen{}_pop{}".format(gen, n): {"gen": gen, "pop": n,
    #                                                "score": i.fitness.values[0],
    #                                                "expr": str(subp(i.expr)),
    #                                                "with_dim": 1 if i.withdim else 0,
    #                                                "dim_is_target_dim": 1 if i.dim in target_dim else 0,
    #                                                "gen_dim": "{}{}".format(gen, 1 if i.withdim else 0),
    #                                                "gen_target_dim": "{}{}".format(gen,
    #                                                                                1 if i.dim in target_dim else 0),
    #                                                "socre_dim": i.fitness.values[0] if i.withdim else 0,
    #                                                "socre_target_dim": i.fitness.values[
    #                                                    0] if i.dim in target_dim else 0,
    #                                                } for n, i in enumerate(population) if i is not None}
    #         data_all.update(datamnist)
    #     random.setstate(rst)
    #     # select_gs the next generation individuals
    #     offspring = toolbox.select_gs(population, len_pop)
    #
    #     # Vary the pool of individuals
    #     offspring = varAnd(offspring, toolbox, cxpb, mutpb)
    #
    #     rst = random.getstate()
    #
    #     # Evaluate the individuals with an invalid fitness
    #     invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
    #     # fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
    #     # fitnesses = parallelize(n_jobs=3, func=toolbox.evaluate, iterable=invalid_ind,  respective=False)
    #     fitnesses = toolbox.parallel(iterable=invalid_ind)
    #     for ind, fit in zip(invalid_ind, fitnesses):
    #         ind.fitness.values = fit[0],
    #         ind.expr = fit[1]
    #         ind.dim = fit[2]
    #         ind.withdim = fit[3]
    #
    #     add_ind = toolbox.select_kbest_target_dim(population, K_best=0.1 * len_pop)
    #     # add_ind2 = toolbox.select_kbest_dimless(population, K_best=0.2 * len_pop)
    #     # add_ind3 = toolbox.select_kbest(population, K_best=5)
    #     # offspring += add_ind
    #     # offspring += add_ind2
    #     # offspring += add_ind3
    #
    #     # Update the hall of fame with the generated individuals
    #     hal = len(add_ind)
    #     if halloffame is not None:
    #         halloffame.update(add_ind)
    #         if len(halloffame.items) > 0 and halloffame.items[-1].fitness.values[0] >= 0.95:
    #             print(halloffame.items[-1])
    #             print(halloffame.items[-1].fitness.values[0])
    #             break
    #
    #     population[:] = offspring
    #     # Append the current generation statistics to the logbook
    #     record = stats.compile(population) if stats else {}
    #     logbook.record(gen=gen, nevals=len(population), **record)
    #     if verbose:
    #         print(logbook.stream)
    #     random.setstate(rst)
    #
    # store = Store()
    # store.to_csv(data_all)
    # return population, logbook


def selKbestDim(pop, K_best=10, dim_type=None, fuzzy=False):
    fit_attr = "fitness"
    chosen = sorted(pop, key=operator.attrgetter(fit_attr))
    chosen.reverse()

    if dim_type is None:
        add_ind = [ind for ind in chosen if ind.withdim == 1]
    elif dim_type is 'integer':
        add_ind = [ind for ind in chosen if ind.y_dim.isinteger]
    elif dim_type is 'ignore':
        add_ind = chosen
    elif isinstance(dim_type, list):
        add_ind = [ind for ind in chosen if ind.y_dim in dim_type]
    elif isinstance(dim_type, Dim):
        if fuzzy:
            add_ind = [ind for ind in chosen if ind.y_dim.is_same_base(dim_type)]
        else:
            add_ind = [ind for ind in chosen if ind.y_dim == dim_type]
    else:
        raise TypeError("dim_type should be None, 'integer', special Dim or list of Dim")
    if K_best is None:
        try:
            K_best = round(len(add_ind) / 10)
        except:
            K_best = 0
    if len(add_ind) >= round(K_best):
        return add_ind[:round(K_best)]
    else:
        return add_ind



#!/usr/bin/python3.7
# -*- coding: utf-8 -*-

# @Time   : 2019/7/29 19:46
# @Author : Administrator
# @Software: PyCharm
# @License: BSD 3-Clause


"""
# Just a copy from xenonpy

eg:
# fig = plt.figure(figsize=[6.4, 4.8])
# # fig.patch.set_facecolor("white")
# ax = fig.add_subplot(111)
# # # ax = fig.add_axes([0.15,0.1,0.7,0.3])
# # ax.patch.set_facecolor("w")
# # # ax.patch.set_alpha(0.5)
# # [ax.spines[_].set_linewidth(3) for _ in ['left', 'right', 'bottom', 'top']]
# # ax.tick_params(direction='in', which='major', width=3, length=7, colors='black', labelsize=15)
# plt.xlabel("x", )
# plt.ylabel("y", )
# plt.title("x_name")
# # plt.xlim()
# # plt.ylim()
# # plt.legend()
# plt.show()

plt.sca(ax1)
plt.sca(ax2)
"""
from itertools import product

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import rcParams
from matplotlib.colorbar import ColorbarBase
from scipy.stats import pearsonr
from sklearn.datasets import load_boston


class BasePlot(object):
    """
    add definition to matpltlib.plt to draw better figures

    """

    def __init__(self, font="Times"):
        if font == "Times":
            rcParams['font.family'] = 'serif'
            rcParams['font.serif'] = ['Times new roman']
        # rcParams['figure.dpi'] = 300
        # rcParams['figure.figsize'] = [6.4, 4.8]
        # rcParams['figure.titlesize'] = 20

        rcParams['axes.titlesize'] = 18
        rcParams['axes.labelsize'] = 16
        rcParams['axes.linewidth'] = 2

        rcParams['xtick.major.width'] = 2
        rcParams['ytick.major.width'] = 2
        rcParams['xtick.major.size'] = 4
        rcParams['ytick.major.size'] = 4

        # rcParams['xtick.labelsize'] = 14
        # rcParams['ytick.labelsize'] = 14
        rcParams['xtick.direction'] = 'in'
        rcParams['ytick.direction'] = 'in'

        rcParams['legend.fontsize'] = 14

        rcParams['axes.grid'] = True
        rcParams['grid.alpha'] = 0.1

    @staticmethod
    def base_axes():
        fig = plt.figure()
        ax = fig.add_subplot(111)
        return ax, plt

    @staticmethod
    def base_figure():
        plt.figure(0)
        return plt

    def yy_numpy(self, y_true, y_predict, strx='y_true', stry='y_predict'):
        return self.scatter(y_true, y_predict, strx=strx, stry=stry)

    @staticmethod
    def bar(data, types=None, labels=None, strx='x', stry='y'):

        if isinstance(data, pd.DataFrame):
            types = list(data.columns.values)
            labels = list(data.index.values)
            data = data.values

        data = np.array(data)
        if data.ndim == 1:
            data.reshape((-1, 1))

        index = np.arange(data.shape[0]) * data.shape[1] // 2
        bar_width = 0.35

        if not labels:
            labels = ["f%s" % i for i in index]
        if not types:
            types = list(range(data.shape[1]))
        opacity = 0.4

        fig, ax = plt.subplots()
        for i, (x, typei) in enumerate(zip(data.T, types)):
            ax.bar(index + i * bar_width, x, bar_width, label=typei, alpha=opacity)

        plt.xlabel(strx)
        plt.ylabel(stry)
        ax.set_xticks(index + len(types) * bar_width / 2 - bar_width / 2)
        ax.set_xticklabels(labels)
        ax.legend()
        # fig.tight_layout()

        sns.boxplot()

    @staticmethod
    def scatter(y_true, y_predict, strx='y_true', stry='y_predict'):
        x, y = y_true, y_predict
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.scatter(x, y, marker='o', s=50, alpha=0.7, c='orange', linewidths=None, edgecolors='blue')
        ax.plot([min(x), max(x)], [min(x), max(x)], '--', ms=5, lw=2, alpha=0.7, color='black')
        plt.xlabel(strx)
        plt.ylabel(stry)

    @staticmethod
    def lines(y, x=None, line_labels=None, strx='x', stry='y', ):
        if isinstance(y, pd.DataFrame):
            line_labels = list(y.columns.values)
            y = y.values
        fig = plt.figure()
        ax = fig.add_subplot(111)
        y = np.array(y)
        if not line_labels or len(line_labels) != y.shape[1]:
            labels = ["f%s" % i for i in range(y.shape[1])]
        else:
            labels = line_labels
        if x is None:
            if y.ndim == 2:
                for i in range(y.shape[1]):
                    ax.plot(y[:, i], lw=1.5, marker='o', label=labels[i])
            if y.ndim == 1:
                ax.plot(y, lw=1.5, marker='o', label=labels[0])
        else:
            x = np.array(x)
            if y.ndim == 2:
                for i in range(y.shape[1]):
                    ax.plot(x, y[:, i], lw=1.5, marker='o', label=labels[i])
            if y.ndim == 1:
                ax.plot(x, y, lw=1.5, marker='o', label=labels[0])
        plt.xlabel(strx)
        plt.ylabel(stry)
        ax.legend()

    @staticmethod
    def line_scatter(x, y_scatter, y_lines, strx='x', stry='y'):

        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.scatter(x, y_scatter, marker='o', s=10, alpha=0.7, c='orange', linewidths=None, edgecolors='blue')
        ax.plot(y_lines, '-', ms=5, lw=2, alpha=0.7, color='black')
        plt.xlabel(strx)
        plt.ylabel(stry)

    @staticmethod
    def corr(data, square=True, linewidths=.5, annot=False):
        fig = plt.figure()
        fig.add_subplot(111)
        # plt.xticks(rotation='90')
        sns.heatmap(data, cmap="seismic", square=square, linewidths=linewidths, annot=annot, xticklabels=True,
                    yticklabels=True)

    @staticmethod
    def violin(strx, stry, data):
        fig = plt.figure()
        fig.add_subplot(111)
        sns.violinplot(x=strx, y=stry, data=data,
                       linewidth=2,  # 线宽
                       width=0.8,  # 箱之间的间隔比例
                       palette='hls',  # 设置调色板
                       order=None,  # 筛选类别
                       scale='area',  # 测度小提琴图的宽度：area-面积相同，count-按照样本数量决定宽度，width-宽度一样
                       gridsize=50,  # 设置小提琴图边线的平滑度，越高越平滑
                       # bw = 0.8        # 控制拟合程度，一般可以不设置
                       hue='smoker',  # 分类
                       split=True,  # 设置是否拆分小提琴图
                       inner="quartile"  # 设置内部显示类型 → “box”, “quartile”, “point”, “stick”, None
                       )

    @staticmethod
    def box(strx, stry, data):
        sns.boxplot(x=strx, y=stry, data=data,
                    linewidth=2,  # 线宽
                    width=0.8,  # 箱之间的间隔比例
                    fliersize=3,  # 异常点大小
                    palette='hls',  # 设置调色板
                    whis=1.5,  # 设置IQR
                    notch=True,  # 设置是否以中值做凹槽
                    order=['Thur', 'Fri', 'Sat', 'Sun'],  # 筛选类别
                    )

        sns.swarmplot(x=strx, y=stry, data=data, color='k', size=3, alpha=0.8)

    @staticmethod
    def yy_jointplot(str_x, str_y, data):
        sns.jointplot(str_x, str_y, data,
                      kind='reg')

    @staticmethod
    def imshow(np_array):
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.imshow(np_array)

    def show(self):
        plt.show()


def lin_cof(x0):
    results_list = []
    xx = x0.T
    yy = x0.T
    for a in xx:
        for b in yy:
            results = pearsonr(a, b)[0]
            results_list.append(results)
    results1 = np.array(results_list).reshape((x0.shape[-1], x0.shape[-1]))
    return results1


def corr_plot(x_cof, x_name=None, left_down=None, right_top=None, threshold_left=0, threshold_right=0.9,
              title="pearsonr coefficient", label_axis="off", front_raito=1):
    """
    plot corr

    Parameters
    ----------
    x_cof:np.ndarray
        correlation coefficient matrix
    x_name:list,None
        feature names
    left_down:None,"pie","fill","text","circle"
        type for left_down
    right_top:None,"pie","fill","text","circle"
        type for right_top
    threshold_left:int
        threshold for show.
    threshold_right:int
        threshold for show.
    title:str
        picture title
    label_axis:"left","right","off"
        label_axis
    front_raito:float
        front scare for show
    """
    x_cof = np.round(x_cof, 2)

    name = x_name

    size = x_cof
    or_size = np.nan_to_num((abs(size) / size) * (1 - abs(size)))

    n = size.shape[0]
    explode = (0, 0)
    gs = gridspec.GridSpec(n, n)
    gs.update(wspace=0, hspace=0)

    cmap = plt.get_cmap("bwr")  # args
    fill_colors = cmap(size / 2 + 0.5)  # args

    fig = plt.figure(figsize=(6, 6), frameon=True)  # args

    title_fontsize = round(15 * front_raito)  # c_args
    ax_fontsize = round(14 * front_raito)
    score_fontsize = round(12 * front_raito)
    circle_size = round(600 * front_raito)

    fig.text(0.5, 0.05, title, fontsize=title_fontsize, horizontalalignment='center',
             verticalalignment='center')  # zou, xia

    for i, j in product(range(n), range(n)):
        if j < i and abs(size[i, j]) >= threshold_left:
            types = left_down
        elif j > i and abs(size[i, j]) >= threshold_right:
            types = right_top
        else:
            types = None

        if types is "pie":
            ax = plt.subplot(gs[i, j])
            ax.pie((size[i, j], or_size[i, j]), explode=explode, labels=None, autopct=None, shadow=False,
                   startangle=90,
                   colors=[fill_colors[i, j], 'w'], wedgeprops=dict(width=1, edgecolor='black', linewidth=0.5),
                   counterclock=False,
                   frame=False, center=(0, 0), )
            ax.set_xlim(-1, 1)
            ax.axis('equal')

        elif types is "fill":
            ax = plt.subplot(gs[i, j])
            ax.set_facecolor(fill_colors[i, j])
            [ax.spines[_].set_color('w') for _ in ['right', 'top', 'left', 'bottom']]

            ax.set_xticks([])
            ax.set_yticks([])

        elif types is "fillandtext":
            ax = plt.subplot(gs[i, j])
            ax.set_facecolor(fill_colors[i, j])
            [ax.spines[_].set_color('w') for _ in ['right', 'top', 'left', 'bottom']]

            ax.text(0.5, 0.5, size[i, j],
                    fontdict={"color": "black"},  # args
                    fontsize=score_fontsize,  # c_arg
                    horizontalalignment='center', verticalalignment='center')
        elif types is "text":
            ax = plt.subplot(gs[i, j])
            ax.text(0.5, 0.5, size[i, j],
                    fontdict={"color": "b"},  # args
                    fontsize=score_fontsize,  # c_arg
                    horizontalalignment='center', verticalalignment='center')
            ax.set_xticks([])
            ax.set_yticks([])
            # plt.axis('off')
        elif types is "circle":
            ax = plt.subplot(gs[i, j])
            ax.axis('equal')
            ax.set_xlim(-1, 1)
            ax.scatter(0, 0, color=fill_colors[i, j], s=circle_size * abs(size[i, j]) ** 2)
            ax.set_xticks([])

            ax.set_yticks([])
            # plt.axis('off')

        else:
            pass

    for k in range(n):
        ax = plt.subplot(gs[k, k])

        # ax.axis('equal')
        # ax.set_xlim(-1, 1)
        # ax.scatter(0, 0, color=fill_colors[k, k], s=circle_size * abs(size[k, k]))
        # ax.set_xticks([])
        #
        # ax.set_yticks([])

        ax.text(0.5, 0.5, name[k], fontsize=ax_fontsize, horizontalalignment='center', verticalalignment='center')
        ax.set_xticks([])
        ax.set_yticks([])
        if label_axis is "left":
            color = ["w", "w", "b", "b"]
            [ax.spines[i].set_color(j) for i, j in zip(['right', 'top', 'left', 'bottom'], color)]
        elif label_axis is "right":
            color = ["b", "b", "w", "w"]
            [ax.spines[i].set_color(j) for i, j in zip(['right', 'top', 'left', 'bottom'], color)]
        else:
            plt.axis('off')

    fig.subplots_adjust(right=0.75)
    cbar_ax = fig.add_axes([0.80, 0.125, 0.03, 0.75])
    ColorbarBase(cbar_ax, cmap=cmap)
    fig.set_size_inches(7, 6, forward=True)
    plt.show()


if __name__ == '__main__':
    data = load_boston(return_X_y=False)

    name0 = data["feature_names"]
    x = data["datamnist"]
    y = data["target"]
    x_cof = np.corrcoef(x.T)

    corr_plot(x_cof, name0, left_down="circle", right_top="text", threshold_right=0.7, label_axis="off",
              front_raito=0.6)



import copy
import sys

import numpy as np
import sympy
from sympy import Number, Expr
from sympy.core.numbers import ComplexInfinity, NumberSymbol

from featurebox.symbol.calculation.coefficient import get_args

"""lambidfy"""


def general_expr_dict(self, expr_init_map, free_symbol, gsym_map, simplifying=False):
    """gen expr"""

    init_sub = expr_init_map.items()
    for i, j in init_sub:
        self = self.xreplace({sympy.Symbol(i): j})

    fs = free_symbol
    func = sympy.lambdify(fs[0], self, modules=(gsym_map, "sympy"))

    res = func(*fs[1])
    if simplifying:
        res = sympy.simplify(res)

    return res


def general_expr(self, pset, simplifying=False):
    """

    Parameters
    ----------
    simplifying: bool
    self:sympy.Expr
    pset:SymbolSet

    Returns
    -------

    """

    res = general_expr_dict(self, pset.expr_init_map, pset.free_symbol,
                            pset.gsym_map, simplifying=simplifying)

    return res


def group_str(self, pset, feature_name=False):
    """
    return expr just build by input feature name.
    Parameters
    ----------
    self:sympy.Expr or SymbolTree
    pset:SymbolSet
    feature_name:Bool

    Returns
    -------

    """
    #####get expr
    if isinstance(self, Expr):
        expr = copy.deepcopy(self)
        expr = simple(expr, pset.gro_ter_con)[0]
    else:
        if hasattr(self, "coef_expr"):
            expr = self.coef_expr
        elif hasattr(self, "expr") and self.expr is not None:
            expr = self.expr
        else:
            if hasattr(self, "capsule"):
                self = self.capsule
            try:
                expr = compile_context(self, pset.context, pset.gro_ter_con)
            except TypeError:
                raise TypeError("the first inupt must be sympy.expr or SymbolTree,"
                                "the second is SymbolSet")

    ### replace newi to (xi+xj*xk...)
    e_map_va1 = list(pset.expr_init_map.items())
    e_map_va1.reverse()
    for i, j in e_map_va1:
        expr = expr.subs(i, j)

    ### to str
    name_subd = str(expr)

    ### replace Vi() to Vi*(),Vi+()
    arg_list = get_args(expr, sole=False)
    V_map1 = {ar.name: str([str(_) for _ in ar.arr.ravel()]) for ar in arg_list if
              hasattr(ar, "arr") and ar.tp == "Coef"}
    V_map2 = {ar.name: str([str(_) for _ in ar.arr.ravel()]) for ar in arg_list if
              hasattr(ar, "arr") and ar.tp == "Const"}

    V_map_va1 = list(V_map1.items())
    V_map_va2 = list(V_map2.items())
    V_map_va1.reverse()
    V_map_va2.reverse()

    for i, j in V_map_va1:
        name_subd = name_subd.replace(i, "%s*" % j)
    for i, j in V_map_va2:
        name_subd = name_subd.replace(i, "%s+" % j)

    ### replace gxi to [xi,xj]
    t_map_va1 = list(pset.terminals_init_map.items())
    t_map_va1.reverse()
    for i, j in t_map_va1:
        name_subd = name_subd.replace(i, j)

    ### replace ci to (float,)
    c_map_va1 = list(pset.data_x_dict.keys())
    c_map_va1 = [i for i in c_map_va1 if "c" in i]
    c_map_va1.reverse()
    for i in c_map_va1:
        name_subd = name_subd.replace(i, "%.3e" % float(pset.data_x_dict[i]))

    ### replace represent xi to (latex,)
    if feature_name:
        if pset.terminals_fea_map:
            for j1, j2 in pset.terminals_fea_map.values():
                name_subd = name_subd.replace(j1, j2)
        else:
            print("Don not assign the feature_name to pset when pest.add_features")
    print(name_subd)
    return name_subd


def simple(expr01, groups):
    """str to sympy.Expr function.
    add conv to MMdd and MMul.
    !!!!!!! the calcualte conv need conform with np_func().
        is_jump: jump the calculate >= 3 (group_size).
        keep: the calculate is return then input group_size or 1.
    """

    def max_method(expr):

        new = [calculate_number(i) for i in expr.args]
        exprarg_new = list(zip(*new))[0]
        n = list(list(zip(*new))[1])
        expr = expr.func(*exprarg_new)
        n.append(1)
        le = len(set(n))
        if le >= 3:
            return expr, np.nan
        else:
            return expr, max(n)

    def calculate_number(expr):

        if isinstance(expr, sympy.Symbol):
            return expr, groups[expr.name]
        elif isinstance(expr, (Number, ComplexInfinity)):
            return expr, 1
        elif isinstance(expr, NumberSymbol):
            return expr, 1
        else:

            if hasattr(expr.func, "keep"):
                expr_arg, ns = calculate_number(expr.args[0])

                if expr.func.keep:  ###["Self,Conv"]
                    if ns == 1:
                        expr = expr_arg
                        return expr, ns
                    elif ns == 2:
                        expr = expr.func(expr_arg)
                        expr.conu = ns
                        return expr, ns
                    elif ns >= 3:
                        if expr.func.is_jump:
                            expr = expr_arg
                            return expr, ns
                        else:
                            expr = expr.func(expr_arg)
                            expr.conu = ns
                            return expr, ns
                    else:
                        expr = expr_arg
                        return expr, ns

                else:
                    if ns == 1:
                        expr = expr_arg
                        return expr, ns
                    elif ns == 2:
                        expr = expr.func(expr_arg)
                        expr.conu = ns
                        return expr, 1
                    elif ns >= 3:
                        if expr.func.is_jump:
                            expr = expr_arg
                            return expr, ns
                        else:
                            expr = expr.func(expr_arg)  ###["MAdd", "MMul"]
                            expr.conu = ns
                            return expr, 1
                    else:
                        # expr = expr_arg
                        expr = expr.func(expr_arg)
                        expr.conu = ns
                        return expr, ns

            elif hasattr(expr, "arr"):
                #### expr, ns = max_method(expr)
                #### assert expr.arr.shape[0] == ns  #
                return max_method(expr)

            else:
                return max_method(expr)

    expr01 = calculate_number(expr01)
    return expr01


def compile_context(expr, context, gro_ter_con, simplify=True):
    """Compile the expression *expr*.

    :param expr: Expression to compile. It can either be a PrimitiveTree,
                 a string of Python code or any object that when
                 converted into string produced a valid Python code
                 expression.
    :param context: dict
    :param simplify: bool
    :param gro_ter_con: list if group_size
    :returns: a function if the primitive set has 1 or more arguments,
              or return the results produced by evaluating the tree.

    """
    if isinstance(expr, str):
        code = expr
    else:
        code = repr(expr)

    try:
        expr = eval(code, context, {})
    except MemoryError:
        _, _, traceback = sys.exc_info()
        raise MemoryError("DEAP : Error in tree evaluation :"
                          " Python cannot evaluate a tree higher than 90. "
                          "To avoid this problem, you should use bloat control on your "
                          "operators. See the DEAP documentation for more information. "
                          "DEAP will now abort.").with_traceback(traceback)
    if simplify:
        expr = simple(expr, gro_ter_con)[0]
    return expr


def compile_(expr, pset):
    """Compile the expression *expr*.

    :param expr: Expression to compile. It can either be a PrimitiveTree,
                 a string of Python code or any object that when
                 converted into string produced a valid Python code
                 expression.
    :param pset: Primitive set against which the expression is compile.
    :returns: a function if the primitive set has 1 or more arguments,
              or return the results produced by evaluating the tree.
    """
    if isinstance(expr, str):
        code = expr
    else:
        code = repr(expr)
    if len(pset.arguments) > 0:
        # This section is a stripped version of the lambdify
        # function of SymPy 0.6.6.
        args = ",".join(arg for arg in pset.arguments)
        code = "lambda {args}: {code}".format(args=args, code=code)
    try:
        return eval(code, pset.context, {})
    except MemoryError:
        _, _, traceback = sys.exc_info()
        raise MemoryError("DEAP : Error in tree evaluation :"
                          " Python cannot evaluate a tree higher than 90. "
                          "To avoid this problem, you should use bloat control on your "
                          "operators. See the DEAP documentation for more information. "
                          "DEAP will now abort.").with_traceback(traceback)



# -*- coding: utf-8 -*-

# @Time   : 2019/4/5 21:30
# @Author : Administrator
# @Project : feature-optimization-mutibest
# @FileName: cluster_and_plot.py
# @Software: PyCharm

"""
cluster the feature and plot
"""

import warnings

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from sklearn.cluster import DBSCAN, KMeans
from sklearn.datasets import load_breast_cancer
from sklearn.metrics import pairwise_distances

warnings.filterwarnings("ignore")


def cluster(data_t, eps=1, dra=False):
    distance_cof = pairwise_distances(data_t, metric='correlation')

    def my_ravel(data_cof):
        for i in range(data_cof.shape[0]):
            for k in range(i + 1, data_cof.shape[0]):
                yield i, k, data_cof[i, k]

    db = DBSCAN(algorithm='auto', eps=eps, metric='precomputed',
                metric_params=None, min_samples=2, n_jobs=None, p=None)
    db.fit(distance_cof)
    # c = db.core_sample_indices_
    label = db.labels_
    set_label = list(set(label))

    group = {"%s" % aim: [i for i in range(len(label)) if label[i] == aim] for aim in set_label}

    km = KMeans(n_clusters=1, random_state=0, tol=1e-3, max_iter=1000)
    core = []
    for i, j in list(group.items())[:-1]:
        if len(j) < 3:
            true_point = j[0]
        else:
            data_j = data_t[j]
            km.fit(data_t[j])
            point = km.cluster_centers_
            if len(point) > 1:
                point = point[0]
            dis = np.argmin(pairwise_distances(data_j, point, metric='correlation'))
            true_point = j[dis]
        core.append(true_point)
    core.sort()
    nodesize = [100] * distance_cof.shape[0]

    for i in core:
        nodesize[i] *= 10
    if dra:
        g = nx.Graph()
        distance_weight = list(my_ravel(distance_cof))
        g.add_weighted_edges_from(distance_weight)
        # edges=nx.get_edge_attributes(g, 'weight').items()
        # edges, weights = zip(*nx.get_edge_attributes(g, 'weight').items())
        edges, weights = [], []
        for _ in range(len(distance_weight)):
            if distance_weight[_][2] < 0.5:
                weights.append(distance_weight[_][2])
                edges.append(distance_weight[_][:2])

        pos = nx.layout.kamada_kawai_layout(g)
        # pos = nx.layout.spectral_layout(g)
        # scaler = MinMaxScaler(feature_range=(0, 1), copy=True)
        # pca = PCA()
        # poss = pca.fit_transform(scaler.fit_transform(data_t.T).T)
        # poss= scaler.fit_transform(poss)[:, 0:2]
        # pos = {i: j for i, j in enumerate(poss)}

        nx.draw(g, pos, labels={i: i for i in core}, edgelist=edges, edge_color=np.around(weights, decimals=1) ** 0.5,
                edge_cmap=plt.cm.Blues_r, edge_labels=nx.get_edge_attributes(g, 'weight'), edge_vmax=0.7,
                node_color=np.array(label) + 1, vmin=0,
                node_cmap=plt.cm.tab20c, node_size=nodesize, width=weights,
                )

        plt.show()
    return group, core


if __name__ == "__main__":
    data = load_breast_cancer().data
    datat = data.T
    cluster(datat, eps=0.3, dra=True)



import unittest
from functools import reduce

import sympy
from sympy.core.numbers import oo as Inf
from sympy.core.numbers import nan

from featurebox.symbol.functions.gsymfunc import NewArray
from featurebox.symbol.functions.gsymfunc import gsym_map as sym_map



class MyTestCase(unittest.TestCase):
    def setUp(self):
        self.a = sympy.Symbol("a")*sympy.Symbol("a")
        self.nan = nan
        self.n = NewArray([1.0, 1,  Inf])
        self.x = sympy.Symbol("x")
        self.k = sympy.Symbol("k")
        self.y = sympy.Symbol("y")
        self.z = sympy.Symbol("z")
        self.arr3 = NewArray([self.y, self.z, self.x])
        self.arr3_ = NewArray([self.x, self.z, self.k])
        self.arr3_NaN = NewArray([self.x, self.z, nan])
        self.arr4 = NewArray([self.x, self.y, self.z, self.k])

    def test_mul(self):
        self.assertEqual(self.a * self.arr3, NewArray([self.y * self.a, self.z * self.a, self.x * self.a]))
        self.assertEqual(self.arr3*self.a , NewArray([self.y * self.a, self.z * self.a, self.x * self.a]))

        arrs = NewArray(
            [i * j for i, j in zip(self.arr3_ , self.arr3)]
        )
        self.assertEqual(self.arr3_ * self.arr3,arrs)

        self.assertNotEqual(self.arr3_NaN * self.arr3,arrs)

        arrs2 = NewArray(
            [i * j for i, j in zip(self.n,self.arr3)]
        )
        self.assertEqual(self.n*self.arr3, arrs2)

    def test_add(self):
        self.assertEqual(self.a + self.arr3, NewArray([self.y + self.a, self.z + self.a, self.x + self.a]))
        self.assertEqual(self.arr3+self.a , NewArray([self.y + self.a, self.z + self.a, self.x + self.a]))

        arrs = NewArray(
            [i + j for i, j in zip(self.arr3_ , self.arr3)]
        )
        self.assertEqual(self.arr3_ + self.arr3,arrs)

        self.assertNotEqual(self.arr3_NaN + self.arr3,arrs)

        arrs2 = NewArray(
            [i + j for i, j in zip(self.n,self.arr3)]
        )
        self.assertEqual(self.arr3 + self.n, arrs2)

    def test_sub(self):
        self.assertEqual(self.arr3 - self.a, NewArray([self.y - self.a, self.z - self.a, self.x - self.a]))

        arrs = NewArray(
            [i - j for i, j in zip(self.arr3_ , self.arr3)]
        )
        self.assertEqual(self.arr3_ - self.arr3,arrs)

        self.assertNotEqual(self.arr3_NaN - self.arr3,arrs)

        arrs2 = NewArray(
            [i - j for i, j in zip(self.n,self.arr3)]
        )
        self.assertEqual(self.n - self.arr3, arrs2)

    def test_rsub(self):
        self.assertEqual(self.a - self.arr3, NewArray([self.a-self.y ,self.a- self.z , self.a-self.x]))
        self.assertEqual(self.a - self.arr3,  -NewArray([self.y - self.a, self.z - self.a, self.x - self.a]))

        arrs = NewArray(
            [i - j for i, j in zip(self.arr3_ , self.arr3)]
        )
        self.assertEqual(self.arr3_ - self.arr3,arrs)

        self.assertNotEqual(self.arr3_NaN - self.arr3,arrs)

        arrs2 = NewArray(
            [i - j for i, j in zip(self.n,self.arr3)]
        )
        self.assertEqual(self.n - self.arr3, arrs2)

    def test_rdiv(self):
        self.assertEqual(self.a / self.arr3, NewArray([self.a/self.y ,self.a/ self.z , self.a/self.x]))
        # self.assertEqual(self.a / self.arr3, 1/NewArray([self.y / self.a, self.z / self.a, self.x / self.a]))

        arrs = NewArray(
            [i / j for i, j in zip(self.arr3_ , self.arr3)]
        )
        self.assertEqual(self.arr3_ / self.arr3,arrs)

        self.assertNotEqual(self.arr3_NaN / self.arr3,arrs)

        arrs2 = NewArray(
            [i / j for i, j in zip(self.n,self.arr3)]
        )
        self.assertEqual(self.n / self.arr3, arrs2)

    def test_div(self):
        self.assertEqual(self.arr3/self.a, NewArray([self.y/self.a ,self.z/ self.a , self.x/self.a]))
        # self.assertEqual(self.arr3/self.a, 1/NewArray([self.a/self.y ,self.a/ self.z , self.a/self.x]))

        arrs = NewArray(
            [i / j for i, j in zip(self.arr3_ , self.arr3)]
        )
        self.assertEqual(self.arr3_ / self.arr3,arrs)

        self.assertNotEqual(self.arr3_NaN / self.arr3,arrs)

        arrs2 = NewArray(
            [i / j for i, j in zip(self.n,self.arr3)]
        )
        self.assertEqual(self.n / self.arr3, arrs2)

    def test_exp(self):
        f = sym_map()["exp"]
        self.assertEqual(f(self.arr3), NewArray([f(self.y),f(self.z) , f(self.x)]))

    def test_sin(self):
        f = sym_map()["exp"]
        self.assertEqual(f(self.arr3), NewArray([f(self.y),f(self.z), f(self.x)]))
    def test_Flat(self):
        f = sym_map()["MAdd"]
        self.assertEqual(f(self.arr3), sum([self.y, self.z, self.x]))
    def test_comp(self):
        f = sym_map()["MMul"]
        self.assertEqual(f(self.arr3), self.y*self.z*self.x)
    def test_conv(self):
        f = sym_map()["Conv"]
        self.assertEqual(f(self.arr3), self.arr3)
        assert True==(f(NewArray([self.y,self.z]))==NewArray([self.z,self.y]))
    def test_msub(self):
        f = sym_map()["MSub"]
        self.assertEqual(f(self.arr3), self.arr3)
        self.assertEqual(f(NewArray([self.y,self.z])), self.y-self.z)
    def test_mdiv(self):
        f = sym_map()["MDiv"]
        self.assertEqual(f(self.arr3), self.arr3)
        self.assertEqual(f(NewArray([self.y,self.z])), self.y/self.z)
    #
    # def test_ppp(self):
    #     syk_map
if __name__ == '__main__':
    unittest.main()


"""
vector coef and vector const, which is a UndefinedFunction to excape the auto calculation of numpy to sympy.
"""
import copy
import warnings
from collections import Counter

import numpy as np
import sympy
from scipy import optimize
from sympy import Function
from sympy.core.function import UndefinedFunction



class Coef(UndefinedFunction):
    """
    generate metaclass, the type of identity is .arr,.tp ,rather isinstance
    """

    def __new__(mcs, name, arr):

        def lfun(x):
            if isinstance(x,np.ndarray):
                return arr * x
            else:
                return arr.ravel() * x

        implementation = lfun
        f = super(Coef,mcs).__new__(mcs, name=name, _imp_=staticmethod(implementation))
        f.arr = arr
        f.name = name
        f.tp = "Coef"
        return f

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, Coef):
            return np.all(np.all((self.arr, other.arr)))
        else:
            return False

    def __hash__(self):
        return hash((self.name, str(self.arr)))


class Const(UndefinedFunction):
    """
    generate metaclass, the type of identity is .arr,.tp ,rather isinstance
    """

    def __new__(mcs, name, arr):

        def lfun(x):
            if isinstance(x,np.ndarray):
                return arr + x
            else:
                return arr.ravel() + x

        implementation = lfun
        f = super(Const,mcs).__new__(mcs, name=name, _imp_=staticmethod(implementation))
        f.arr = arr
        f.name = name
        f.tp = "Const"
        return f

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, Coef):
            return np.all(np.all((self.arr, other.arr)))
        else:
            return False

    def __hash__(self):
        return hash((self.name, str(self.arr)))


def get_args(expr, sole=True):
    """

    Parameters
    ----------
    expr:sympy.Expr
    sole:only find unique term

    Returns
    -------
    list
    """

    def _get_args(expr_):
        """"""
        list_arg = []
        for argi in expr_.args:
            list_arg.append(argi)
            if argi.args:
                re = _get_args(argi)
                list_arg.extend(re)

        return list_arg

    list_a = _get_args(expr)
    if sole:
        count = Counter(list_a)
        term = [i for i, j in count.items() if j == 1]
        list_a = term

    return list_a


def find_args(expr_, patten):
    """find the term of patten, judge by hash rather type"""
    if len(expr_.args) > 0:
        if patten in expr_.args:
            return expr_.args
        else:
            for argi in expr_.args:
                d = find_args(argi, patten)
                if d is not None:
                    return d


def replace_args(expr_, old, new):
    """find the term of patten, judge by hash rather type"""
    if len(expr_.args) > 0:
        if old in expr_.args:
            compo = list(expr_.args)
            indexs = compo.index(old)
            compo.remove(old)
            compo.insert(indexs, new)
            return expr_.func(*compo)
        else:

            compo = [replace_args(argi, old, new) for argi in expr_.args]
            return expr_.func(*compo)
    else:
        return expr_


def _replace_args_first(expr_, old, new, keep=False):
    """find the term of patten, judge by hash rather type"""
    if not keep:

        if len(expr_.args) > 0:

            if old in expr_.args:
                compo = list(expr_.args)
                indexs = compo.index(old)
                compo.remove(old)
                compo.insert(indexs,new)
                return expr_.func(*compo),True
            else:
                compo = list(expr_.args)
                for i, argi in enumerate(expr_.args):
                    argi, keep = _replace_args_first(argi, old, new)
                    if keep:
                        compo[i] = argi
                        break
                    else:
                        pass
                return expr_.func(*compo),True

        else:
            return expr_, False
    else:
        return expr_, keep


def replace_args_first(expr_, old, new):
    """a"""
    return _replace_args_first(expr_, old, new)[0]


def add_coefficient(expr01, inter_add=True, inner_add=False, vector_add=False):
    """
    try add placeholder coefficient to sympy expression.
    add to Wi,A,B normal coefficient to expression.
    add V, Vi vecotr coefficient to expression, for this type of coefficent ,
    there should be expr01.conu for Function("MAdd"), Function("MSub").
    more detial can be found in ..translate.simple

    Parameters
    ----------
    expr01: Expr
        sympy expressions
    inter_add: bool
        bool
    inner_add: bool
        bool
    vector_add:bool
        bool

    Returns
    -------
    expr
    """
    cof_list = []
    cof_dict = {}
    if isinstance(expr01, sympy.Add):
        wiss = [sympy.Symbol("W%s" % i) for i, j in enumerate(expr01.args)]
        args_new = [(wi, sympy.Mul(wi, ei)) if not ei.is_number else (None, ei) for ei, wi in zip(expr01.args, wiss)]
        wis, we = zip(*args_new)
        cof_list.extend([wi for wi in wis if wi is not None])
        argss = sympy.Add(*we)
        expr01 = argss

    elif isinstance(expr01, (Function("MAdd"), Function("MSub"))):

        exprin1 = expr01.args[0]
        conu = expr01.conu
        if isinstance(exprin1, sympy.Add):
            wiss = [sympy.Symbol("W%s" % i) for i, j in enumerate(exprin1.args)]
            args_new = [(wi, sympy.Mul(wi, ei)) if not ei.is_number else (None, ei) for ei, wi in
                        zip(exprin1.args, wiss)]
            wis, we = zip(*args_new)
            cof_list.extend([wi for wi in wis if wi is not None])
            argss = sympy.Add(*we)
            expr01 = expr01.func(argss)

        if vector_add:
            if conu > 1:
                Wi = sympy.Symbol("V")
                arg = expr01.args[0]
                expr02 = expr01.func(sympy.Mul(Wi, arg))
                cof_dict[Wi] = conu
                expr01 = expr02
        else:
            A = sympy.Symbol("A")
            expr01 = sympy.Mul(expr01, A)
            cof_list.append(A)

    else:
        A = sympy.Symbol("A")
        expr01 = sympy.Mul(expr01, A)
        cof_list.append(A)

    if inner_add:

        arg_list = get_args(expr01)

        arg_list = [i for i in arg_list if i not in expr01.args]
        cho = []
        cho_add = [i.args for i in arg_list if isinstance(i, sympy.Add)]
        cho_add = [[_ for _ in cho_addi if not _.is_number] for cho_addi in cho_add]
        [cho.extend(i) for i in cho_add]

        a_cho = [sympy.Symbol("k%s" % i) for i in range(len(cho))]

        for ai, choi in zip(a_cho, cho):
            expr02 = expr01.xreplace({choi: sympy.Mul(ai, choi)})
            cof_list.append(ai)
            expr01 = expr02

        if vector_add:
            cho_add2 = [i for i in arg_list if isinstance(i, (Function("MAdd"), Function("MSub"))) if
                        hasattr(i, "conu") and i.conu > 1]

            for i, j in enumerate(cho_add2):

                Wi = sympy.Symbol("V%s" % i)
                arg = j.args[0]
                arg_new = j.func(sympy.Mul(Wi, arg))
                expr02 = expr01.xreplace({j: arg_new})
                cof_dict[Wi] = j.conu
                expr01 = expr02

    if inter_add:
        B = sympy.Symbol("B")
        expr01 = sympy.Add(expr01, B)
        cof_list.append(B)

    return expr01, cof_list, cof_dict


class CheckCoef(object):
    """
    group the coef and pack the calculate part out of loop.
    """

    def __init__(self, cof_list, cof_dict):
        """

        Parameters
        ----------
        cof_list: Sized
        cof_dict:dict
        """
        self.cof_list = cof_list
        self.cof_dict = cof_dict
        self.cof_dict_keys = list(cof_dict.keys())
        self.cof_dict_values = list(cof_dict.values())
        lt = []
        lt.extend(cof_list)
        lt.extend(list(self.cof_dict_keys))
        self.name = lt
        self.num = len(cof_list) + sum(list(self.cof_dict_values))

    def __len__(self):
        return len(self.name)

    @property
    def ind(self):
        lsa = list(range(len(self.cof_list)))
        n = len(lsa)
        for k in self.cof_dict_values:
            lsi = list(range(k))
            lsi = [lsii + n for lsii in lsi]
            lsa.append(lsi)
            n = lsi[-1] + 1

        return lsa

    def group(self, p, decimals=False):
        """change the p to grpup"""
        p = np.array(p)
        ls = []
        for i in self.ind:
            if isinstance(i, int):
                ls.append(p[i])
            else:
                ps = p[i].reshape((-1, 1))
                ls.append(ps)

        if decimals:
            return self.dec(ls)
        else:
            return ls

    def dec(self,ls):
        cof_ = []
        for a_listi, cofi in zip(self.name, ls):

            if not isinstance(cofi, np.ndarray):
                cof_.append(float("%.3e" % cofi))
            else:
                cof_.append(np.array([float("%.3e" % i) for i in cofi]).reshape((-1, 1)))
        return cof_


def try_add_coef(expr01, x, y, terminals,
                 filter_warning=True, inter_add=True, inner_add=False, vector_add=False, np_maps=None):
    """
    try calculate predict y by sympy expression with coef.
    if except error return expr self.
    Parameters
    ----------
    vector_add: bool
        add vector coefficent or not
    expr01: sympy.Expr
        sympy expressions
    x: list of np.ndarray
        list of xi
    y: np.ndarray
        y value
    terminals: list of sympy.Symbol
        features and constants
    filter_warning: bool
        bool
    inter_add: bool
        bool
    inner_add: bool
        bool
    np_maps: Callable
        user np.ndarray function

    Returns
    -------
    pre_y:
        np.array or None
    expr01: Expr
        New expr.
    """
    if filter_warning:
        warnings.filterwarnings("ignore")

    expr00 = copy.deepcopy(expr01)

    expr01, a_list, a_dict = add_coefficient(expr01, inter_add=inter_add, inner_add=inner_add, vector_add=vector_add)

    cc = CheckCoef(a_list, a_dict)

    ter = []
    ter.extend(terminals)
    ter.extend(cc.name)

    try:

        func0 = sympy.utilities.lambdify(ter, expr01, modules=[np_maps, "numpy", "math"])

        def func(x_, p):
            """"""
            num_list = []
            num_list.extend(x_)
            p = cc.group(p)
            num_list.extend(p)

            return func0(*num_list)

        def res(p, x_, y_):
            """"""
            ress = y_ - func(x_, p)
            return ress

        result = optimize.least_squares(res, x0=[1.0] * cc.num, args=(x, y), xtol=1e-4, ftol=1e-5,gtol=1e-5,# long
                                        jac='3-point', loss='linear')
        cof = result.x

        cof = cc.group(cof)
        pre_y = func0(*x+cof)
        cof = cc.dec(cof)

        for ai, choi in zip(cc.name, cof):
            if ai in cc.cof_dict_keys:

                fun = Coef(ai.name, choi)
                # replace the Vi to Vi()
                olds0 = find_args(expr01, ai)
                if olds0 is None:
                    raise KeyError("0*wi is 0,and make the placeholder fade out")
                olds = [old for old in olds0 if old is not ai]
                olds = sympy.Mul(*olds)
                expr01 = expr01.xreplace({sympy.Mul(ai, olds): fun(olds)})
            else:
                expr01 = expr01.xreplace({ai: choi})

    except (ValueError, KeyError, NameError, TypeError, ZeroDivisionError):

        expr01 = expr00
        pre_y =None

    return pre_y, expr01



# -*- coding: utf-8 -*-

# @Time    : 2019/11/12 14:21
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause
"""
All part are copy from deap (https://github.com/DEAP/deap)
"""
import copy
import functools
import random
import warnings

import sympy

warnings.filterwarnings("ignore")


class FixedTerminal(object):
    """

    """
    __slots__ = ('name', 'value', 'conv_fct', 'arity')

    def __init__(self, terminal):
        self.value = terminal
        self.name = str(terminal)
        self.conv_fct = str
        self.arity = 0

    def format(self):
        return self.conv_fct(self.value)

    def __eq__(self, other):
        if type(self) is type(other):
            return all(getattr(self, slot) == getattr(other, slot) for slot in self.__slots__)
        else:
            return NotImplemented

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        return self.name

    __repr__ = __str__


class FixedPrimitive(object):
    """

    """
    __slots__ = ('name', 'arity', 'args', 'seq')

    def __init__(self, name, arity):
        self.name = name
        self.arity = arity
        self.args = []
        args = ", ".join(map("{{{0}}}".format, list(range(self.arity))))
        self.seq = "{name}({args})".format(name=self.name, args=args)

    def format(self, *args):
        return self.seq.format(*args)

    def __eq__(self, other):
        if type(self) is type(other):
            return all(getattr(self, slot) == getattr(other, slot) for slot in self.__slots__)
        else:
            return NotImplemented

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        return self.name

    __repr__ = __str__


class FixedSet(object):
    """

    """

    def __init__(self, name):
        self.terminals = []
        self.primitives = []
        self.terms_count = 0
        self.prims_count = 0
        self.arguments = []
        self.context = {"__builtins__": None}
        self.dimtext = {"__builtins__": None}
        self.mapping = dict()
        self.name = name

    def addPrimitive(self, primitive, arity, name=None):

        if name is None:
            name = primitive.__name__

        prim = FixedPrimitive(name, arity)

        assert name not in self.context, "Primitives are required to have a unique x_name. " \
                                         "Consider using the argument 'x_name' to rename your " \
                                         "second '%s' primitive." % (name,)

        self.primitives.append(prim)
        self.context[prim.name] = primitive
        self.prims_count += 1

    def addTerminal(self, terminal, name=None, dim=None):

        if name is None and callable(terminal):
            name = str(terminal)

        assert name not in self.context, "Terminals are required to have a unique x_name. " \
                                         "Consider using the argument 'x_name' to rename your " \
                                         "second %s terminal." % (name,)

        if name is not None:
            self.context[name] = terminal

        prim = FixedTerminal(terminal)
        self.terminals.append(prim)
        self.terms_count += 1
        self.dimtext[name] = dim

    @property
    def terminalRatio(self):
        """Return the ratio of the number of terminals on the number of all
        kind of primitives.
        """
        return self.terms_count / float(self.terms_count + self.prims_count)


class FixedTree(list):
    """

    """
    hasher = str

    def __init__(self, content):
        list.__init__(self, content)

        assert sum(_.arity - 1 for _ in self.primitives) + 1 >= len(self.terminals)
        assert len(self.terminals) >= 2

    @property
    def root(self):
        """

        Returns
        -------
        start site number of tree
        """
        len_ter = len(self.terminals) - 1
        num_pri = list(range(len(self.primitives)))
        num_pri.reverse()
        i = 0
        for i in num_pri:
            if len_ter == 0:
                break
            elif len_ter <= 0:
                raise ("Add terminals or move back the {}".format(self[i - 1]),
                       "because the {} have insufficient terminals, "
                       "need {},but get {}".format(self[i - 1], self[i - 1].arity, len_ter - self[i - 1].arity)
                       )
            len_ter = len_ter - self[i].arity + 1
        if self[i].arity == 1:
            return i
        else:
            return i + 1

    def __deepcopy__(self, memo):

        new = self.__class__(self)
        new.__dict__.update(copy.deepcopy(self.__dict__, memo))
        return new

    def __hash__(self):
        return hash(self.hasher(self))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __str__(self):
        return self.str_run_index(self.root)

    __repr__ = __str__

    @property
    def pri_site(self):
        return tuple([p for p, primitive in enumerate(self) if primitive.arity >= 1 and p >= self.root])

    @property
    def ter_site(self):
        return tuple([t for t, primitive in enumerate(self) if primitive.arity == 0])

    @property
    def primitives(self):
        """Return primitives that occur in the expression tree."""
        return [primitive for primitive in self if primitive.arity >= 1]

    @property
    def terminals(self):
        """Return terminals that occur in the expression tree."""
        return [primitive for primitive in self if primitive.arity == 0]

    @classmethod
    def search_y_name(cls, name):
        """

        Parameters
        ----------
        name

        Returns
        -------

        """
        list_name = []
        for i in range(len(cls)):
            if cls[i].name == name:
                list_name.append(i)
        return list_name

    @property
    def number_branch(self):
        """

        Returns
        -------
        dict,a site tree
        """

        def cal():
            coup_dict = {}
            coup = []
            for _ in range(pri_i.arity):
                coup.append(terminals.pop())
            coup.reverse()
            coup_dict[len_pri - i - 1] = coup
            terminals_new.append(coup_dict)

        primitives = self.primitives
        primitives.reverse()
        len_pri = len(primitives)
        terminals = list(range(len_pri, len(self.terminals) + len_pri))
        terminals_new = []
        for i, pri_i in enumerate(primitives):
            if len(terminals) >= pri_i.arity:
                cal()
            else:
                terminals_new.reverse()
                terminals.extend(terminals_new)
                terminals_new = []
                if len(terminals) >= pri_i.arity:
                    cal()
                else:
                    break
        result = terminals_new or terminals
        return result[0]

    def number_branch_index(self, index):
        """
        Returns
        -------
        number_branch for given index
        """
        if index < self.root or index > len(self.primitives):
            raise IndexError("not a primitives index")
        else:
            def run_index(number_branch=None):
                if number_branch is None:
                    number_branch = self.number_branch
                jk = list(number_branch.keys())[0]
                ji = list(number_branch.values())[0]
                if jk == index:
                    return number_branch
                else:
                    repr1 = []
                    for jii in ji:
                        if isinstance(jii, dict):
                            repr1 = run_index(jii)
                        else:
                            repr1 = []
                        if repr1:
                            break
                    return repr1
        set1 = run_index()
        # set1.sort()
        return set1

    def str_run(self, number_branch=None):
        """
        Returns
        -------
        str of tree by given number_branch,default is the str of root number_branch
        """
        if number_branch is None:
            number_branch = self.number_branch
        # print(number_branch)
        args = []
        jk = list(number_branch.keys())[0]
        ji = list(number_branch.values())[0]

        for jii in ji:
            if isinstance(jii, dict):
                repr1 = self.str_run(jii)
            else:
                repr1 = self[jii].name
            args.append(repr1)
        repr1 = self[jk].format(*args)

        return repr1

    def str_run_index(self, i):
        return self.str_run(self.number_branch_index(i))

    def indexs_in_node(self, i):
        """

        Returns
        -------
        indexs in node branch
        """

        def run_index(number_branch=None):
            if number_branch is None:
                number_branch = self.number_branch
            jk = list(number_branch.keys())[0]
            ji = list(number_branch.values())[0]
            sub_index = []
            for jii in ji:
                if isinstance(jii, dict):
                    repr1 = run_index(jii)
                else:
                    repr1 = [jii]
                sub_index.extend(repr1)
            sub_index.append(jk)

            return sub_index

        res = run_index(number_branch=self.number_branch_index(i))
        res = list(set(res))
        res.sort()
        return res


def FixedSetFill(x_name, rep_name=None, power_categories=None, categories=("Add", "Mul", "Abs", "exp"),
                 partial_categories=None, self_categories=None, dim=None, max_=5,
                 definate_operate=None, definate_variable=None, operate_linkage=None, variable_linkage=None):
    """

    Parameters
    ----------
    :type partial_categories: double list
    partial_categories = [["Add","Mul"],["x4"]]
    :type power_categories: list
    index_categories=[0.5,1,2,3]
    :type dim: list,tuple
    :type x_name: list,tuple
    :type max_
    :type rep_name: list,tuple
    :type categories: list,tuple
    :param self_categories:
    def rem(a):
        return 1-a
    self_categories = [[rem, 1, 'rem']]
    :type definate_variable: list,tuple
    definate_variable = [(-1, [1, ]), ]
    :type definate_operate: list,tuple
    definate_operate = [(-1, [0, ]), ]
    :param operate_linkage
    :param variable_linkage
    """

    def Div(left, right):
        return left / right

    def Sub(left, right):
        return left - right

    def zeroo(_):
        return 0

    def oneo(_):
        return 1

    def rem(a):
        return 1 - a

    def self(_):
        return _

    functions2 = {"Add": sympy.Add, 'Sub': Sub, 'Mul': sympy.Mul, 'Div': Div}
    functions1 = {"sin": sympy.sin, 'cos': sympy.cos, 'exp': sympy.exp, 'log': sympy.ln,
                  'Abs': sympy.Abs, "Neg": functools.partial(sympy.Mul, -1.0),
                  "Rec": functools.partial(sympy.Pow, e=-1.0),
                  'Zeroo': zeroo, "Oneo": oneo, "Rem": rem, "Self": self}

    pset0 = FixedSet('main')

    if power_categories:
        for j, i in enumerate(power_categories):
            pset0.addPrimitive(functools.partial(sympy.Pow, e=i), arity=1, name='pow%s' % j)

    for i in categories:
        if i in functions1:
            pset0.addPrimitive(functions1[i], arity=1, name=i)
        if i in functions2:
            pset0.addPrimitive(functions2[i], arity=2, name=i)

    if partial_categories:
        for partial_categoriesi in partial_categories:
            for i in partial_categoriesi[0]:
                for j in partial_categoriesi[1]:
                    if i in ["Mul", "Add"]:
                        pset0.addPrimitive(functools.partial(functions2[i], sympy.Symbol(j)), arity=1,
                                           name="{}_{}".format(i, j))
                    elif i in ["Div", "Sub"]:
                        pset0.addPrimitive(functools.partial(functions2[i], right=sympy.Symbol(j)), arity=1,
                                           name="{}_{}".format(i, j))
                    else:
                        pass
    if self_categories:
        for i in self_categories:
            pset0.addPrimitive(i[0], i[1], i[2])

    # define terminal
    if not rep_name:
        rep_name = ['{}{}'.format("x", i) for i, _ in enumerate(x_name)]

    rep_name = [sympy.Symbol(i) for i in rep_name]

    if dim is None:
        dim = [None] * len(rep_name)

    assert len(dim) == len(x_name) == len(rep_name)

    for sym, dimi in zip(rep_name, dim):
        pset0.addTerminal(sym, name=str(sym), dim=dimi)

    dict_pri = dict(zip([_.name for _ in pset0.primitives], range(len(pset0.primitives))))
    dict_ter = dict(zip([_.name for _ in pset0.terminals], range(len(pset0.terminals))))

    if max_ is None:
        max_ = len(pset0.terminals)

    # define limit
    def link_check(checking_linkage):
        """"""
        if checking_linkage is None:
            checking_linkage = [[]]
        assert isinstance(checking_linkage, (list, tuple))
        if not isinstance(checking_linkage[0], (list, tuple)):
            checking_linkage = [checking_linkage, ]
        return checking_linkage

    operate_linkage = link_check(operate_linkage)
    variable_linkage = link_check(variable_linkage)

    operate_linkage = [[j - max_ for j in i] for i in operate_linkage]
    linkage = operate_linkage + variable_linkage

    if definate_operate:
        definate_operate = [list(i) for i in definate_operate]
        for i, j in enumerate(definate_operate):
            j = list(j)
            definate_operate[i][1] = [dict_pri[_] if _ in dict_pri else _ for _ in j[1]]
        err_str = [i for j in definate_operate for i in j[1] if isinstance(i, str)]
        if len(err_str) > 0:
            raise NameError("{} in definate_operate sholud be defined in categories".format(set(err_str)))
    if definate_variable:
        definate_variable = [list(i) for i in definate_variable]
        for i, j in enumerate(definate_variable):
            j = list(j)
            definate_variable[i][1] = [dict_ter[_] if _ in dict_ter else _ for _ in j[1]]

    pset0.definate_operate = definate_operate
    pset0.definate_variable = definate_variable
    pset0.linkage = linkage

    pset0.rep_name_list = rep_name
    pset0.real_name_list = x_name
    pset0.dim_list = dim

    pset0.max_ = max_
    print(dict_pri)
    print(dict_ter)

    return pset0


def generate_index(pset, min_=None, max_=None):
    """

    Parameters
    ----------
    pset
    min_
    max_

    Returns
    -------

    """
    if max_ is None:
        max_ = len(pset.terminals_and_constant)
    _ = min_
    min_ = max_

    pri2 = [i for i in pset.primitives if i.arity == 2]
    pri1 = [i for i in pset.primitives if i.arity == 1]

    max_varibale_set_long = max_
    varibale_set_long = random.randint(min_, max_varibale_set_long)
    '''random'''
    trem_set = random.sample(pset.terminals_and_constant, varibale_set_long) * 20
    '''sequence'''
    # trem_set = pset.terminals[:varibale_set_long] * 20

    init_operator_long = max_varibale_set_long * 4
    individual1 = []
    for i in range(init_operator_long):
        trem = random.choice(pri2) if random.random() > 0.5 * len(pri1) / len(
            pset.primitives) else random.choice(pri1)
        individual1.append(trem)
    individual2 = []
    for i in range(varibale_set_long):
        trem = trem_set[i]
        individual2.append(trem)
    # define protect primitives
    pri2 = [i for i in pset.primitives if i.arity == 2]
    protect_individual = []
    for i in range(varibale_set_long):
        trem = random.choice(pri2)
        protect_individual.append(trem)

    definate_operate = pset.definate_operate
    definate_variable = pset.definate_variable
    linkage = pset.linkage

    if definate_operate:
        for i in definate_operate:
            individual1[i[0]] = pset.primitives[random.choice(i[1])]

    if definate_variable:
        for i in definate_variable:
            individual2[i[0]] = pset.terminals_and_constant[random.choice(i[1])]

    individual_all = protect_individual + individual1 + individual2
    if linkage:
        for i in linkage:
            for _ in i:
                individual_all[_] = individual_all[i[0]]

    return individual_all


def cxOnePoint_index(ind1, ind2, pset):
    """

    Parameters
    ----------
    ind1
    ind2
    pset

    Returns
    -------

    """
    linkage = pset.linkage
    root = max(ind1.root, ind2.root)
    index = random.randint(root, root + len(ind1.pri_site))
    ind10 = copy.copy(ind1)
    ind20 = copy.copy(ind2)
    ind10[index:] = ind2[index:]
    ind20[index:] = ind1[index:]
    if linkage:
        for i in linkage:
            for _ in i:
                ind10[_] = ind10[i[0]]
                ind20[_] = ind20[i[0]]
    return ind10, ind20


def mutUniForm_index(ind10, pset, ):
    """

    Parameters
    ----------
    ind1
    pset

    Returns
    -------

    """
    linkage = pset.linkage
    definate_operate = pset.definate_operate
    ranges = list(range(ind10.pri_site[-1] + 1))
    pri2 = [i for i in pset.primitives if i.arity == 2]
    pri1 = [i for i in pset.primitives if i.arity == 1]
    index = random.choice(ind10.pri_site)
    ind10[index] = random.choice(pri2) if random.random() > 0.5 * len(pri1) / len(
        pset.primitives) else random.choice(pri1)

    if definate_operate:
        for i in definate_operate:
            ind10[ranges[i[0]]] = pset.primitives[random.choice(i[1])]

    if linkage:
        for i in linkage:
            for _ in i:
                ind10[_] = ind10[i[0]]

    return ind10,



# -*- coding: utf-8 -*-

# @Time    : 2019/11/6 16:20
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause

import numpy as np
from sklearn.utils._random import check_random_state

'''
    """please copy the code to test dont use by import"""

    X = init(n_feature=10, m_sample=100)  # 样本数

    X_name = ["X%s" % i for i in range(X.shape[1])]
    for i, X_i in enumerate(X_name):
        locals()[X_i] = X[:, i]

    # 添加关系
    """relation"""

    X2 = X0+X1

    """noise"""

    X0 = add_noise(X0,ratio=0.01)


    X0 = add_noise(X0, ratio=0.2)
    X3 = add_noise(X0 / X2,ratio=0.2)
    X4 = add_noise(X0 * X2,ratio=0.2)

    """重定义"""
    X_all = [eval("X%s" % i) for i in range(X.shape[1])]
    X_new = np.vstack(X_all).T

    """定义函数"""  # 改变函数
    y = X0 ** 3 + X2
    return X, y
'''


def init(m_sample=100, n_feature=10):
    n = n_feature
    m = m_sample
    mean = [1] * n
    cov = np.zeros((n, n))
    for i in range(cov.shape[1]):
        cov[i, i] = 1
    rdd = check_random_state(1)
    X = rdd.multivariate_normal(mean, cov, m)
    return X


def add_noise(s, ratio):
    print(s.shape)
    rdd = check_random_state(1)
    return s + rdd.random_sample(s.shape) * np.max(s) * ratio


from sklearn.utils import Bunch


class CAMData(object):
    def __init__(self, samples, shape_x, random_state=None, method="poisson", inter=True, cut_zero=True,
                 re_range=None,
                 question="reg", rank=True):
        """

        Parameters
        ----------
        samples:int
        shape_x:int or tuple of int
        random_state:None or int
        method:"poission" or None
        inter:bool
            integer or not.
        cut_zero:
            non-negative or not
        re_range:None or tuple of int, default is (0,1)
            range
        question:"reg" or "clf"
        rank:bool
            rank values for y, just for the "reg"
        """
        if isinstance(shape_x, int):
            self.shape_x = (shape_x,)
        else:
            self.shape_x = shape_x

        self.rdd = check_random_state(random_state)
        self.random_state = random_state
        self.samples = samples
        self.method = method
        self.inter = inter
        self.cut_zero = cut_zero
        self.re_range = re_range
        self.question = question
        self.rank = rank

    def _gen_X(self):
        rdd = self.rdd
        method = self.method
        inter = self.inter
        cut_zero = self.cut_zero
        re_range = self.re_range

        if method == "poission":

            data = rdd.poisson(lam=1.0, size=(self.samples, *self.shape_x))
        else:
            data = rdd.random_sample(size=(self.samples, *self.shape_x))

        if re_range is None:
            pass
        elif re_range == "zeroone":
            a, b = 0, 1
            data = (b - a) * data + a
        else:
            a, b = re_range[0], re_range[1]
            data = (b - a) * data + a
        if inter:
            data = np.round(data)
        if cut_zero:
            data[np.where(data < 0)] = 0
        return data

    def gen_x(self, featurize_x=True, self_func=None):
        """

        Parameters
        ----------
        featurize_x: bool or "total" or "each"
            deal with x.
            if True or "each", the self_func should be a func to deal with each xi sample.
            input parameters are (xi,i,random_state)
            such as:

            def self_func(xi,i,random_state):
                return xi[0,0,0]+=i+random_state.randint(7,11,1), "random add"

            if "total", the self_func should be a func to deal with x data.
            input parameters are (self,x)
            such as:

            def featurized_x(self, x)
                return x*=2, "x*2"

            if False, pass
        self_func : callable object,or None,or str, default is None
            1.None: return xi and ""
            2.str:"cp_channel_to3","de_2Dsite_i"
            3.the function to deal to xi, return xi_new and feature messagei
            4.or the function to deal to x, return x_new and feature message

        """
        if self_func is None:
            self._featurized_xi = self._no_featurize_xi
        elif self_func is "cp_channel_to3":
            self._featurized_xi = self.cp_channel_to3
        elif self_func is "de_2Dsite_i":
            self._featurized_xi = self.de_2Dsite_i
        elif self_func is "de_2Dsite":
            self._featurized_xi = self.de_2Dsite
        else:
            raise TypeError(
                "self_func is callable object,or None,or 'cp_pipeline_to3','de_2Dsite','de_2Dsite_i', default is None")

        x = self._gen_X()

        if not featurize_x:
            self.feature = None
        elif featurize_x is True or featurize_x is "each":
            x, self.feature = self.featurized_x(x)
        elif featurize_x is "total":
            self.featurized_x = self._featurized_xi
            x, self.feature = self.featurized_x(x)
        else:
            raise TypeError('featurize_x is bool or "total" or "each"')

        return x

    @staticmethod
    def de_2Dsite_i(xi, i, random_state):
        """
        one instance just for 2d data
        a manipulation on xi and get new xi and feature message"""
        c = random_state.randint(7, 11, 1) / 10
        # c = random_state.random(1)
        b_a = i / c
        a = random_state.randint(0, b_a + 1, size=1)
        b = b_a + a

        ainx, ainy = random_state.randint(0, 10, size=2)
        ain = ainx, ainy
        bin = ainx + 5, ainy + 7
        cin = ainx + 2, ainy + 5
        xi[ain] = a
        xi[bin] = b
        xi[cin] = c
        featurei = ((ain, bin, cin), (a, b, c))
        # print(xi[cin]*(xi[bin]-xi[ain]))

        return xi, featurei

    def de_2Dsite(self, x):
        """resort the x"""

        def site(xi, i, rd):
            a, b, c = rd.random_sample(3)
            ainx, ainy = rd.randint(2, 10, size=2)
            ain = ainx, ainy
            ratio = rd.randint(10, 15, size=1) / 10
            bin = int(ainx + 13 * ratio), int(ainy + 1 * ratio)
            cin = int(ainx + 7 * ratio), int(ainy + 14 * ratio)
            xi[ain] = a
            xi[bin] = b
            xi[cin] = c
            featurei = ((ain, bin, cin), (b - a) * c)
            return xi, featurei

        if self._featurized_xi is None:
            return x, ""
        else:
            rd = self.rdd
            s1, *ss = x.shape
            x_new = []
            feature = []
            rank = []
            for i in range(s1):
                xi = x[i]
                xi, featurei = site(xi, i, rd)
                xi = np.expand_dims(xi, axis=0)
                x_new.append(xi)
                feature.append(featurei)
                rank.append(featurei[1])
            index = np.argsort(rank)
            x_new = np.array(x_new)[index]
            feature = np.array(feature)[index]
            x_new = np.concatenate(x_new, axis=0)
            return x_new, feature

    @staticmethod
    def cp_channel_to3(x):
        x = np.expand_dims(x, axis=-1)
        x = np.repeat(x, 3, axis=-1)

        return x, "dim (sample, ... , 3)"

    @staticmethod
    def cp_channel_to1(x):
        x = np.expand_dims(x, axis=-1)
        x = np.repeat(x, 1, axis=-1)

        return x, "dim (sample, ... , 1)"

    @staticmethod
    def _no_featurize_xi(xi, *kwargs):

        raise NotImplementedError("no self function")

    def featurized_x(self, x):
        if self._featurized_xi is None:
            return x, ""
        else:
            rd = self.rdd
            s1, *ss = x.shape
            x_new = []
            feature = []
            for i in range(s1):
                xi = x[i]
                xi, featurei = self._featurized_xi(xi, i, rd)
                xi = np.expand_dims(xi, axis=0)
                x_new.append(xi)
                feature.append(featurei)

            x_new = np.concatenate(x_new, axis=0)
            return x_new, feature

    def gen_y(self):
        question = self.question
        rank = self.rank
        shape_y = self.samples
        rdd = self.rdd
        if question is "reg":
            if rank:
                target = np.array(list(range(shape_y))) / shape_y * (1 + rdd.random_sample(shape_y) * 0.1)
                target = np.reshape(target, (shape_y, 1))
            else:
                target = rdd.random_sample(size=(shape_y, 1))
        else:
            if rank:
                target1 = np.zeros((int(shape_y / 2), 1))
                target2 = np.ones((shape_y - int(shape_y / 2), 1))
                target = np.concatenate((target1, target2), axis=0)
            else:
                target = rdd.randint(0, 2, size=(shape_y, 1))
        return target

    def gen_data(self, featurize_x=False, self_func=None):
        x = self.gen_x(featurize_x=featurize_x, self_func=self_func)
        y = self.gen_y()
        x = x.astype(np.float32)
        y = y.astype(np.float32)
        return Bunch(x=x, y=y, feature=self.feature)


if __name__ == "__main__":
    sample = 5000

    camdata = CAMData(15, 15, random_state=1, question="reg", inter=False, re_range=(-0.5, 1))
    data = camdata.gen_data(featurize_x=False)
    x = data.x
    y = data.y
    x = x.astype(np.float32)

    # t_x = trans_to_tensor(x)



# -*- coding: utf-8 -*-

# @Time   : 2019/7/13 19:27
# @Author : Administrator
# @Project : feature_toolbox
# @FileName: test_sum.py
# @Software: PyCharm

"""
USM
this is a union select method for feature subsets.
key:
1.gather performance of different model
2.raise the best one from similar learning performance subsets
3.rank the raised subsets and penalty the size of subsets

node == feature subset
"""
import numpy as np

from featurebox.selection.ugs import UGS, displacement


def _kamada_kawai_solve(dist_mtx, dim=2):
    # Anneal node locations based on the Kamada-Kawai cost-function,
    # using the supplied matrix of preferred inter-node distances,
    # and starting locations.
    from networkx import circular_layout
    pos = circular_layout(list(range(dist_mtx.shape[0])), dim=dim)
    pos_arr = np.array([pos[n] for n in list(range(dist_mtx.shape[0]))])

    from scipy.optimize import minimize

    meanwt = 1e-3
    costargs = (np, 1 / (dist_mtx + np.eye(dist_mtx.shape[0]) * 1e-3),
                meanwt, dim)

    optresult = minimize(_kamada_kawai_costfn, pos_arr.ravel(),
                         method='Newton-CG', args=costargs, jac=True)

    return optresult.x.reshape((-1, dim))


def _kamada_kawai_costfn(pos_vec, np, invdist, meanweight, dim):
    # Cost-function and gradient for Kamada-Kawai layout algorithm
    nNodes = invdist.shape[0]
    pos_arr = pos_vec.reshape((nNodes, dim))

    delta = pos_arr[:, np.newaxis, :] - pos_arr[np.newaxis, :, :]
    nodesep = np.linalg.norm(delta, axis=-1)
    direction = np.einsum('ijk,ij->ijk',
                          delta,
                          1 / (nodesep + np.eye(nNodes) * 1e-3))

    offset = nodesep * invdist - 1.0
    offset[np.diag_indices(nNodes)] = 0

    cost = 0.5 * np.sum(offset ** 2)
    grad = (np.einsum('ij,ij,ijk->ik', invdist, offset, direction) -
            np.einsum('ij,ij,ijk->jk', invdist, offset, direction))

    # Additional parabolic term to encourage mean position to be near origin:
    sumpos = np.sum(pos_arr, axis=0)
    cost += 0.5 * meanweight * np.sum(sumpos ** 2)
    grad += meanweight * sumpos

    return cost, grad.ravel()


class SUM(UGS):
    """

    """

    def __init__(self, estimator, slices, estimator_n=None, n_jobs=2, batch_size=1):
        """

        Parameters
        ----------
        estimator : list
            list of base estimator or GridSearchCV from sklearn
        slices: list
            the lists of the index of feature subsets, each feature subset is a node,each int is the index of X
            Examples 3 nodes
            [[1,4,5],[1,4,6],[1,2,7]]
        estimator_n: list
            default indexes of estimator
        """
        super().__init__(estimator, slices, estimator_n=estimator_n, n_jobs=n_jobs, batch_size=batch_size)

    @staticmethod
    def cal_r(kk):
        r_matrix = []
        for i in kk:
            r_matrix.append(np.sqrt(np.sum((kk - i) ** 2, axis=1)))
        return np.array(r_matrix)

    def distance_method(self):
        cal_binary_distance_all_model = [self.cal_binary_distance_all(estimator_i=i) for i in self.estimator_n]
        score_all_model = [self.cv_score_all(estimator_i=i) for i in self.estimator_n]
        max_node = [[np.argmax(_)] for _ in score_all_model]
        print("best node for different model", max_node)
        long = len(score_all_model[0])
        iter_ = np.linspace(np.min(np.array(cal_binary_distance_all_model)),
                            np.max(np.array(cal_binary_distance_all_model)), num=100 * long)

        # unify = [np.max(_) for _ in cal_binary_distance_all_model]
        # cal_binary_distance_all_model = [i / j for i, j in zip(cal_binary_distance_all_model, unify)]

        rank = []
        distance = []
        for steps in iter_:
            circle = set(list(range(long)))
            for dis_all, score_all, max_node_i in zip(cal_binary_distance_all_model, score_all_model, max_node):
                circle_i = set(np.where(dis_all[:, max_node_i[0]] <= steps)[0])
                circle &= circle_i
            addd = list(circle - set(rank))
            if addd:
                rank.extend(list(circle - set(rank)))
                distance.extend([steps] * len(addd))
        slices_rank = [str(self.slices[i]) for i in rank]
        print("select slices %s" % slices_rank[0], "select node %s" % rank[0])
        return list(zip(rank, slices_rank, distance))

    def distance_kk_method(self):
        cal_binary_distance_all_model1 = [self.cal_binary_distance_all(estimator_i=i) for i in self.estimator_n]
        score_all_model = [self.cv_score_all(estimator_i=i) for i in self.estimator_n]

        cal_binary_distance_all_model1 = [displacement(i) for i in cal_binary_distance_all_model1]
        KK_dis_all_model = [_kamada_kawai_solve(_, dim=2) for _ in cal_binary_distance_all_model1]
        cal_binary_distance_all_model = [self.cal_r(_) for _ in KK_dis_all_model]

        # unify = [np.max(_) for _ in cal_binary_distance_all_model]
        # cal_binary_distance_all_model = [i / j for i, j in zip(cal_binary_distance_all_model, unify)]

        max_node = [[np.argmax(_)] for _ in score_all_model]
        print("best node for different model", max_node)

        long = len(score_all_model[0])
        iter_ = np.linspace(np.min(np.array(cal_binary_distance_all_model)),
                            np.max(np.array(cal_binary_distance_all_model)), num=100 * long)

        rank = []
        distance = []
        for steps in iter_:
            circle = set(list(range(long)))
            for dis_all, score_all, max_node_i in zip(cal_binary_distance_all_model, score_all_model, max_node):
                circle_i = set(np.where(dis_all[:, max_node_i[0]] <= steps)[0])
                circle &= circle_i
            addd = list(circle - set(rank))
            if addd:
                rank.extend(list(circle - set(rank)))
                distance.extend([steps] * len(addd))
        slices_rank = [str(self.slices[i]) for i in rank]
        print("select slices %s" % slices_rank[0], "select node %s" % rank[0])

        return list(zip(rank, slices_rank, distance))

    def y_distance_method(self):
        cal_y_distance_all_model = [self.cal_y_distance_all(estimator_i=i) for i in self.estimator_n]

        max_node = [[np.argmin(_)] for _ in cal_y_distance_all_model]

        print(max_node)

        long = len(cal_y_distance_all_model[0])
        iter_ = np.linspace(np.min(np.array(cal_y_distance_all_model)),
                            np.max(np.array(cal_y_distance_all_model)), num=100 * long)
        rank = []
        distance = []
        for steps in iter_:
            circle = set(list(range(long)))
            for dis_all in cal_y_distance_all_model:
                circle_i = set(np.where(dis_all <= steps)[0])
                circle &= circle_i
            addd = list(circle - set(rank))
            if addd:
                rank.extend(list(circle - set(rank)))
                distance.extend([steps] * len(addd))
        slices_rank = [str(self.slices[i]) for i in rank]

        print("select slices %s" % slices_rank[0], "select node %s" % rank[0])

        return list(zip(rank, slices_rank, distance))

    def cal_binary_add_y_distance_all(self):

        cal_binary_distance_all_model = [self.cal_binary_distance_all(estimator_i=i) for i in self.estimator_n]

        cal_y_distance_all_model = [self.cal_y_distance_all(estimator_i=i) for i in self.estimator_n]

        matrix = [np.concatenate((i, j.reshape(-1, 1)), axis=1)
                  for i, j in zip(cal_binary_distance_all_model, cal_y_distance_all_model)]

        cal_y_distance_all_model0 = [np.append(i, 0) for i in cal_y_distance_all_model]

        matrix = [np.concatenate((i, j.reshape(1, -1)), axis=0)
                  for i, j in zip(matrix, cal_y_distance_all_model0)]

        # unify = [np.max(_) for _ in matrix]
        # cal_binary_distance_all_model = [i / j for i, j in zip(matrix, unify)]

        return matrix

    def y_distance_kk_method(self):
        matrix = self.cal_binary_add_y_distance_all()
        matrix = [displacement(i) for i in matrix]
        KK_dis_all_model = [_kamada_kawai_solve(_, dim=2) for _ in matrix]
        cal_binary_distance_all_model = [self.cal_r(_) for _ in KK_dis_all_model]

        cal_y_distance_all_model = [i[:-1, -1] for i in cal_binary_distance_all_model]
        # max_node = [[np.argmin(_)] for _ in cal_y_distance_all_model]

        long = len(cal_y_distance_all_model[0])
        iter_ = np.linspace(np.min(np.array(cal_y_distance_all_model)),
                            np.max(np.array(cal_y_distance_all_model)), num=100 * long)
        rank = []
        distance = []
        for steps in iter_:
            circle = set(list(range(long)))
            for dis_all in cal_y_distance_all_model:
                circle_i = set(np.where(dis_all <= steps)[0])
                circle &= circle_i
            addd = list(circle - set(rank))
            if addd:
                rank.extend(list(circle - set(rank)))
                distance.extend([steps] * len(addd))
        slices_rank = [str(self.slices[i]) for i in rank]

        print("select slices %s" % slices_rank[0], "select node %s" % rank[0])

        return list(zip(rank, slices_rank, distance))

    @staticmethod
    def _pareto(y, sign=None):

        m = y.shape[0]
        n = y.shape[1]
        if not sign:
            sign = np.array([1] * n)
        y = y * sign
        front_point = []
        for i in range(m):
            data_new = y[i, :].reshape(1, -1) - y
            data_max = np.max(data_new, axis=1)
            data_in = np.min(data_max)
            if data_in >= 0:
                front_point.append(i)
        # print("pareto_method point number:", len(front_point))
        front_point = np.array(front_point)

        return front_point

    def pareto_method(self, sign=None):
        y = np.array([self.cv_score_all(estimator_i=i) for i in self.estimator_n]).T

        max_node = np.argmax(y, axis=0)
        print(max_node)

        front_point = self._pareto(y, sign=sign)

        mean_ = np.mean(y, axis=1)
        mean_front = mean_[front_point]
        rank_ = np.argsort(mean_front)[::-1]
        front_point_rank = front_point[rank_]
        std_front_rank = mean_front[rank_]

        slices_rank = [str(self.slices[i]) for i in front_point_rank]

        print("select slices %s" % slices_rank[0], "select node %s" % front_point_rank[0])

        return list(zip(front_point_rank, slices_rank, std_front_rank))

    def mean_max_method(self):
        score_all = np.array([self.cv_score_all(estimator_i=i) for i in self.estimator_n]).T
        score = np.mean(score_all, axis=1)
        rank = np.argsort(score)[::-1]
        score_ranked = score[rank]
        slices_rank = [str(self.slices[i]) for i in rank]

        print("select slices %s" % slices_rank[0], "select node %s" % rank[0])
        return list(zip(rank, slices_rank, score_ranked))



# #!/usr/bin/python
# # -*- coding: utf-8 -*-
#
# # @Time    : 2019/11/12 15:13
# # @Email   : 986798607@qq.com
# # @Software: PyCharm
# # @License: GNU Lesser General Public License v3.0

import copy
import operator
import os
import time

import sympy
from deap.base import Fitness
from deap.tools import HallOfFame, Logbook
from numpy import random
from sklearn.datasets import load_boston
from sklearn.metrics import r2_score

from featurebox.symbol.base import CalculatePrecisionSet
from featurebox.symbol.base import SymbolSet
from featurebox.symbol.base import SymbolTree
from featurebox.symbol.calculation.translate import compile_context, general_expr
from featurebox.symbol.functions.dimfunc import Dim, dless
from featurebox.symbol.gp import cxOnePoint, varAnd, genGrow, staticLimit, selKbestDim, \
    selTournament, Statis_func, mutUniform, mutShrink, varAndfus, \
    mutDifferentReplacementVerbose, mutNodeReplacementVerbose, selBest, genFull
from featurebox.tools import newclass
from featurebox.tools.exports import Store
from featurebox.tools.packbox import Toolbox


class BaseLoop(Toolbox):
    """Base loop"""

    def __init__(self, pset, pop=500, gen=20, mutate_prob=0.5, mate_prob=0.8, hall=1, re_hall=None,
                 re_Tree=None, initial_min=None, initial_max=3, max_value=5,
                 scoring=(r2_score,), score_pen=(1,), filter_warning=True, cv=1,
                 add_coef=True, inter_add=True, inner_add=False, vector_add=False,
                 cal_dim=False, dim_type=None, fuzzy=False, n_jobs=1, batch_size=40,
                 random_state=None, stats=None, verbose=True,
                 tq=True, store=False, personal_map=False, stop_condition=None):
        """

        Parameters
        ----------
        pset:SymbolSet
            the feature x and traget y and others should have been added.
        pop:int
            number of popolation
        gen:int
            number of generation
        mutate_prob:float
            probability of mutate
        mate_prob:float
            probability of mate(crossover)
        initial_max:int
            max initial size of expression when first producing.
        initial_min : None,int
            max initial size of expression when first producing.
        max_value:int
            max size of expression
        hall:int,>=1
            number of HallOfFame(elite) to maintain
        re_hall:None or int>=2
            Notes: only vaild when hall
            number of HallOfFame to add to next generation.
        re_Tree: int
            number of new features to add to next generation.
            0 is false to add.
        personal_map:bool or "auto"
            "auto" is using premap and with auto refresh the premap with individual.\n
            True is just using constant premap.\n
            False is just use the prob of terminals.
        scoring: list of Callbale, default is [sklearn.metrics.r2_score,]
            See Also sklearn.metrics
        score_pen: tuple of  1, -1 or float but 0.
            >0 : max problem, best is positive, worse -np.inf
            <0 : min problem, best is negative, worse np.inf
            Notes:
            if multiply score method, the scores must be turn to same dimension in preprocessing
            or weight by score_pen. Because the all the selection are stand on the mean(w_i*score_i)
            Examples: [r2_score] is [1],
        cv=int,sklearn.model_selection._split._BaseKFold
            default =1, means not cv
        filter_warning:bool
            filter warning or not
        add_coef:bool
            add coef in expression or not.
        inter_add：bool
            add intercept constant or not
        inner_add:bool
            dd inner coeffcients or not
        n_jobs:int
            default 1, advise 6
        batch_size:int
            default 40, depend of machine
        random_state:int
            None,int
        cal_dim:bool
            excape the dim calculation
        dim_type:Dim or None or list of Dim
            "coef": af(x)+b. a,b have dimension,f(x) is not dnan. \n
            "integer": af(x)+b. f(x) is interger dimension. \n
            [Dim1,Dim2]: f(x) in list. \n
            Dim: f(x) ~= Dim. (see fuzzy) \n
            Dim: f(x) == Dim. \n
            None: f(x) == pset.y_dim
        fuzzy:bool
            choose the dim with same base with dim_type,such as m,m^2,m^3.
        stats:dict
            details of logbook to show. \n
            Map:\n
            values 
                = {"max": np.max, "mean": np.mean, "min": np.mean, "std": np.std, "sum": np.sum}
            keys
                = {\n
                   "fitness": just see fitness[0], \n
                   "fitness_dim_max": max problem, see fitness with demand dim,\n
                   "fitness_dim_min": min problem, see fitness with demand dim,\n
                   "dim_is_target": demand dim,\n
                   "coef":  dim is true, coef have dim, \n
                   "integer":  dim is integer, \n
                   ...
                   }
            if stats is None, default is :\n
                stats = {"fitness_dim_max": ("max",), "dim_is_target": ("sum",)}   for cal_dim=True
                stats = {"fitness": ("max",)}                                      for cal_dim=False
            if self-definition, the key is func to get attribute of each ind./n
            Examples:
                def func(ind):\n
                    return ind.fitness[0]
                stats = {func: ("mean",), "dim_is_target": ("sum",)}
        verbose:bool
            print verbose logbook or not
        tq:bool
            print progress bar or not
        store:bool or path
            bool or path
        stop_condition:callable
            stop condition on the best ind of hall, which return bool,the true means stop loop.
            Examples:
                def func(ind):\n
                    c = ind.fitness.values[0]>=0.90
                    return c
        """
        super(BaseLoop, self).__init__()
        assert initial_max <= max_value, "the initial size of expression should less than max_value limitation"
        if cal_dim:
            assert all(
                [isinstance(i, Dim) for i in pset.dim_ter_con.values()]), \
                "all import dim of pset should be Dim object."

        random.seed(random_state)

        self.max_value = max_value
        self.pop = pop
        self.gen = gen
        self.mutate_prob = mutate_prob
        self.mate_prob = mate_prob
        self.verbose = verbose
        self.cal_dim = cal_dim
        self.re_hall = re_hall
        self.re_Tree = re_Tree
        self.store = store
        self.data_all = []
        self.personal_map = personal_map
        self.stop_condition = stop_condition

        self.cpset = CalculatePrecisionSet(pset, scoring=scoring, score_pen=score_pen,
                                           filter_warning=filter_warning, cal_dim=cal_dim,
                                           add_coef=add_coef, inter_add=inter_add, inner_add=inner_add,
                                           vector_add=vector_add, cv=cv,
                                           n_jobs=n_jobs, batch_size=batch_size, tq=tq,
                                           fuzzy=fuzzy, dim_type=dim_type,
                                           )

        Fitness_ = newclass.create("Fitness_", Fitness, weights=score_pen)
        self.PTree = newclass.create("PTrees", SymbolTree, fitness=Fitness_)
        # def produce
        if initial_min is None:
            initial_min = 2
        self.register("genGrow", genGrow, pset=self.cpset, min_=initial_min, max_=initial_max+1,
                      personal_map=self.personal_map)
        self.register("genFull", genFull, pset=self.cpset, min_=initial_min, max_=initial_max+1,
                      personal_map=self.personal_map)
        self.register("gen_mu", genGrow, min_=1, max_=2+1, personal_map=self.personal_map)
        # def selection

        self.register("select", selTournament, tournsize=2)

        self.register("selKbestDim", selKbestDim,
                      dim_type=self.cpset.dim_type, fuzzy=self.cpset.fuzzy)
        self.register("selBest", selBest)

        self.register("mate", cxOnePoint)
        # def mutate

        self.register("mutate", mutUniform, expr=self.gen_mu, pset=self.cpset)

        self.decorate("mate", staticLimit(key=operator.attrgetter("height"), max_value=2 * (max_value+1)))
        self.decorate("mutate", staticLimit(key=operator.attrgetter("height"), max_value=2 * (max_value+1)))

        if stats is None:
            if cal_dim:
                stats = {"fitness_dim_max": ("max",), "dim_is_target": ("sum",)}
            else:
                stats = {"fitness": ("max",)}

        self.stats = Statis_func(stats=stats)
        logbook = Logbook()
        logbook.header = ['gen'] + (self.stats.fields if self.stats else [])
        self.logbook = logbook

        if hall is None:
            hall = 1
        self.hall = HallOfFame(hall)

        if re_hall is None:
            self.re_hall = None
        else:
            if re_hall is 1 or re_hall is 0:
                print("re_hall should more than 1")
                re_hall = 2
            assert re_hall >= hall, "re_hall should more than hall"
            self.re_hall = HallOfFame(re_hall)

    def varAnd(self, *arg, **kwargs):
        return varAnd(*arg, **kwargs)

    def to_csv(self, data_all):
        if self.store:
            if isinstance(self.store, str):
                path = self.store
            else:
                path = os.getcwd()
            file_new_name = "_".join((str(self.pop), str(self.gen),
                                      str(self.mutate_prob), str(self.mate_prob),
                                      str(time.time())))
            try:
                st = Store(path)
                st.to_csv(data_all, file_new_name)
                print("store data to ", path, file_new_name)
            except (IOError, PermissionError):
                st = Store(os.getcwd())
                st.to_csv(data_all, file_new_name)
                print("store data to ", os.getcwd(), file_new_name)

    def maintain_halls(self, population):

        if self.re_hall is not None:
            maxsize = max(self.hall.maxsize, self.re_hall.maxsize)

            if self.cal_dim:
                inds_dim = self.selKbestDim(population, maxsize)
            else:
                inds_dim = self.selBest(population, maxsize)

            self.hall.update(inds_dim)
            self.re_hall.update(inds_dim)

            sole_inds = [i for i in self.re_hall.items if i not in inds_dim]
            inds_dim.extend(sole_inds)
        else:
            if self.cal_dim:
                inds_dim = self.selKbestDim(population, self.hall.maxsize)
            else:
                inds_dim = self.selBest(population, self.hall.maxsize)

            self.hall.update(inds_dim)
            inds_dim = []

        inds_dim = copy.deepcopy(inds_dim)
        return inds_dim

    def re_add(self):
        if self.hall.items and self.re_Tree:
            it = self.hall.items
            indo = it[random.choice(len(it))]
            ind = copy.deepcopy(indo)
            inds = ind.depart()
            if not inds:
                pass
            else:
                inds = [self.cpset.calculate_detail(indi) for indi in inds]
                le = min(self.re_Tree, len(inds))
                indi = inds[random.choice(le)]
                self.cpset.add_tree_to_features(indi)

    def re_fresh_by_name(self, *arr):
        re_name = ["mutate", "genGrow", "genFull"]
        if len(arr) > 0:
            re_name.extend(arr)
        self.refresh(re_name, pset=self.cpset)

    def run(self):
        # 1.generate###################################################################
        population = [self.PTree(self.genFull()) for _ in range(self.pop)]

        for gen_i in range(1, self.gen + 1):

            population_old = copy.deepcopy(population)

            # 2.evaluate###############################################################
            invalid_ind_score = self.cpset.parallelize_score(population_old)

            for ind, score in zip(population_old, invalid_ind_score):
                ind.fitness.values = tuple(score[0])
                ind.y_dim = score[1]
                ind.dim_score = score[2]

            population = population_old

            # 3.log###################################################################
            # 3.1.log-print##############################

            record = self.stats.compile(population) if self.stats else {}
            self.logbook.record(gen=gen_i, **record)
            if self.verbose:
                print(self.logbook.stream)

            # 3.2.log-store##############################
            if self.store:
                datas = [{"gen": gen_i, "name": str(gen_i), "value": str(gen_i.fitness.values),
                          "dimension": str(gen_i.y_dim),
                          "dim_score": str(gen_i.dim_score)} for gen_i in population]
                self.data_all.extend(datas)

            # 3.3.log-hall###############################
            inds_dim = self.maintain_halls(population)

            # 4.refresh################################################################
            # 4.1.re_update the premap ##################
            if self.personal_map is "auto":
                [self.cpset.premap.update(indi, self.cpset) for indi in inds_dim]

            # 4.2.re_add_tree and refresh pset###########
            if self.re_Tree:
                self.re_add()

            self.re_fresh_by_name()

            # 5.break#######################################################
            if self.stop_condition is not None:
                if self.stop_condition(self.hall.items[0]):
                    break
            # 6.next generation#######################################################
            # selection and mutate,mate
            population = self.select(population, len(population) - len(inds_dim))

            offspring = self.varAnd(population, self, self.mate_prob, self.mutate_prob)

            offspring.extend(inds_dim)
            population[:] = offspring

        # 7.store#####################################################################

        if self.store:
            self.to_csv(self.data_all)
        self.hall.items = [self.cpset.calculate_detail(indi) for indi in self.hall.items]

        return self.hall


class MutilMutateLoop(BaseLoop):
    """
    multiply mutate method.
    """

    def __init__(self, *args, **kwargs):
        """See also BaseLoop"""
        super(MutilMutateLoop, self).__init__(*args, **kwargs)

        self.register("mutate0", mutNodeReplacementVerbose, pset=self.cpset, personal_map=self.personal_map)

        self.register("mutate1", mutUniform, expr=self.gen_mu, pset=self.cpset)
        self.decorate("mutate1", staticLimit(key=operator.attrgetter("height"), max_value=2 * self.max_value))

        self.register("mutate2", mutShrink, pset=self.cpset)

        self.register("mutate3", mutDifferentReplacementVerbose, pset=self.cpset, personal_map=self.personal_map)

    def varAnd(self, population, toolbox, cxpb, mutpb):
        names = self.__dict__.keys()
        import re
        patt = r'mutate[0-9]'
        pattern = re.compile(patt)
        result = [pattern.findall(i) for i in names]
        att_name = []
        for i in result:
            att_name.extend(i)

        self.re_fresh_by_name(*att_name)

        fus = [getattr(self, i) for i in att_name]

        off = varAndfus(population, toolbox, cxpb, mutpb, fus)

        return off


class OnePointMutateLoop(BaseLoop):
    """
    limitation height of population, just use mutNodeReplacementVerbose method.
    """

    def __init__(self, *args, **kwargs):
        """See also BaseLoop"""
        super(OnePointMutateLoop, self).__init__(*args, **kwargs)

        self.register("mutate0", mutNodeReplacementVerbose, pset=self.cpset, personal_map=self.personal_map)

        self.register("mutate3", mutDifferentReplacementVerbose, pset=self.cpset, personal_map=self.personal_map)

    def varAnd(self, population, toolbox, cxpb, mutpb):
        names = self.__dict__.keys()
        import re
        patt = r'mutate[0-9]'
        pattern = re.compile(patt)
        result = [pattern.findall(i) for i in names]
        att_name = []
        for i in result:
            att_name.extend(i)

        self.re_fresh_by_name(*att_name)

        fus = [getattr(self, i) for i in att_name]

        off = varAndfus(population, toolbox, cxpb, mutpb, fus)

        return off


class DimForceLoop(MutilMutateLoop):
    """Force select the individual with target dim for next generation"""

    def __init__(self, *args, **kwargs):
        """See also BaseLoop"""
        super(DimForceLoop, self).__init__(*args, **kwargs)
        assert self.cal_dim == True, "For DimForceLoop type, the 'cal_dim' must be True"

        self.register("select", selKbestDim,
                      dim_type=self.cpset.dim_type, fuzzy=self.cpset.fuzzy, force_number=True)

if __name__ == "__main__":
    # data
    data = load_boston()
    x = data["data"]
    y = data["target"]
    c = [6, 3, 4]
    # unit
    from sympy.physics.units import kg

    x_u = [kg] * 13
    y_u = kg
    c_u = [dless, dless, dless]

    x, x_dim = Dim.convert_x(x, x_u, target_units=None, unit_system="SI")
    y, y_dim = Dim.convert_xi(y, y_u)
    c, c_dim = Dim.convert_x(c, c_u)

    z = time.time()

    # symbolset
    pset0 = SymbolSet()
    pset0.add_features(x, y, x_dim=x_dim, y_dim=y_dim, x_group=[[1, 2], [3, 4], [5, 6]])
    pset0.add_constants(c, c_dim=c_dim, c_prob=None)
    pset0.add_operations(power_categories=(2, 3, 0.5),
                         categories=("Add", "Mul", "Sub", "Div", "exp", "Abs"))

    # a = time.time()
    bl = MutilMutateLoop(pset=pset0, gen=4, pop=10, hall=2, batch_size=40, re_hall=2,
                         n_jobs=1, mate_prob=1, max_value=10, initial_max=3,
                         mutate_prob=0.8, tq=True, dim_type="coef",
                         re_Tree=2, store=False, random_state=1,
                         stats={"fitness_dim_max": ["max"], "dim_is_target": ["sum"], "height": ["mean"]},
                         add_coef=True, cal_dim=True, inner_add=False, vector_add=True, personal_map=False)
    # b = time.time()
    bl.run()
    population = [bl.PTree(bl.genFull()) for _ in range(30)]
    pset = bl.cpset
    for i in population:
        # i.ppprint(bl.cpset)
        # i = "exp(gx0/gx1)"

        i = compile_context(i, pset.context, pset.gro_ter_con, simplify=False)
        # print(i)
        # print(i)
        # fun = Coef("V", np.array([1.4,1.3]))
        # i = fun(i)
        # f = Function("MAdd")
        # i = f(i)
        try:
            # group_str(i,pset)
            # i=general_expr(i, pset, simplifying=True)
            i = general_expr(i, pset, simplifying=False)
            # print(i)
            # print(i)
            # pprint(i)
        except NotImplementedError as e:
            print(e)
    # c = time.time()
    # print(c - b, b - a, a - z)
    a, b, c = sympy.Symbol("a"), sympy.Symbol("b"), sympy.Symbol("c")
    print(sympy.simplify((a + (b + 1)) * c) == sympy.simplify(a * c + b * c + c))



# -*- coding: utf-8 -*-

# @Time    : 2019/12/2 19:21
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause
import numpy as np
from numpy.linalg import matrix_rank
from sklearn.utils import check_array


def dimension_check(x, y=None):
    if y is not None:
        x.append(y)
    x = np.array(x).T
    x = check_array(x, ensure_2d=True)
    x = x.astype(np.float64)
    det = matrix_rank(x)
    che = []
    for i in range(x.shape[1]):
        x_new = np.delete(x, i, 1)
        det2 = matrix_rank(x_new)
        che.append(det - det2)
    sum(che)

    if sum(che) == 0:
        return 1
    else:
        return 0



# -*- coding: utf-8 -*-

# @Time    : 2019/10/16 10:39
# @Email   : 986798607@qq.ele_ratio
# @Software: PyCharm
# @License: BSD 3-Clause

import math

import numpy as np


class SDbw(object):
    """
    score the cluster
    this part is copy from https://github.com/zhangsj1007/Machine-Learning/blob/master/S_Dbw.py
    method source:
    Halkidi, M., Batistakis, Y., & Vazirgiannis, M. (2002). Clustering validity checking methods: part II.
    ACM Sigmod Record, 31(3), 19-27.
    """

    def __init__(self, data_cl, data_cluster_ids, center_idxs=None):
        """

        Parameters
        ----------
        data_cl: np.ndarray
            each row is a variable
        center_idxs : np.ndarray
            index of cluster center
        data_cluster_ids : np.ndarray
            label of cluster
        """
        self.data = data_cl
        self.dataClusterIds = data_cluster_ids

        if center_idxs is not None:
            self.centerIdxs = center_idxs
        else:
            self.__getCenterIdxs()

        # spath.center_idxs = center_idxs

        self.clusterNum = len(self.centerIdxs)

        self.stdev = self.__getStdev()

        self.clusterDensity = []

        for i in range(self.clusterNum):
            self.clusterDensity.append(self.__density(self.data[self.centerIdxs[i]], i))

    def __getCenterIdxs(self):
        """ calculate center index """

        self.centerIdxs = []

        clusterDataMp = {}
        clusterDataIdxsMp = {}

        for i in range(len(self.data)):
            entry = self.data[i]
            clusterId = self.dataClusterIds[i]
            clusterDataMp.setdefault(clusterId, []).append(entry)
            clusterDataIdxsMp.setdefault(clusterId, []).append(i)

        for clusterId in sorted(clusterDataMp.keys()):
            subData = clusterDataMp[clusterId]
            subDataIdxs = clusterDataIdxsMp[clusterId]

            m = len(subData[0])
            n = len(subData)

            meanEntry = [0.0] * m

            for entry in subData:
                meanEntry += entry

            meanEntry /= n

            minDist = float("inf")

            centerIdx = 0

            for i in range(len(subData)):
                entry = subData[i]
                idx = subDataIdxs[i]
                dist = self.__dist(entry, meanEntry)
                if minDist > dist:
                    centerIdx = idx
                    minDist = dist

            self.centerIdxs.append(centerIdx)

    def __getStdev(self):
        stdev = 0.0

        for i in range(self.clusterNum):
            varMatrix = np.var(self.data[self.dataClusterIds == i], axis=0)
            stdev += math.sqrt(np.dot(varMatrix.T, varMatrix))

        stdev = math.sqrt(stdev) / self.clusterNum

        return stdev

    def __density(self, center, cluster_idx):

        density = 0

        clusterData = self.data[self.dataClusterIds == cluster_idx]
        for i in clusterData:
            if self.__dist(i, center) <= self.stdev:
                density += 1

        return density

    def __Dens_bw(self):

        variance = 0
        if self.clusterNum == 1:
            return 0
        else:
            for i in range(self.clusterNum):
                for j in range(self.clusterNum):
                    if i == j:
                        continue
                    center = self.data[self.centerIdxs[i]] + self.data[self.centerIdxs[j]] / 2
                    interDensity = self.__density(center, i) + self.__density(center, j)
                    try:
                        variance += interDensity / max(self.clusterDensity[i], self.clusterDensity[j])
                    except ZeroDivisionError:
                        variance += 0
            return variance / (self.clusterNum * (self.clusterNum - 1))

    def __Scater(self):
        thetaC = np.var(self.data, axis=0)
        thetaCNorm = math.sqrt(np.dot(thetaC.T, thetaC))

        thetaSumNorm = 0

        for i in range(self.clusterNum):
            clusterData = self.data[self.dataClusterIds == i]
            theta_ = np.var(clusterData, axis=0)
            thetaNorm_ = math.sqrt(np.dot(theta_.T, theta_))
            thetaSumNorm += thetaNorm_

        return (1 / self.clusterNum) * (thetaSumNorm / thetaCNorm)

    @staticmethod
    def __dist(entry1, entry2):
        return np.linalg.norm(entry1 - entry2)  # 计算data entry的欧拉距离

    def result(self):
        """ return result"""
        return self.__Dens_bw() + self.__Scater()



import operator
import unittest

from sklearn import preprocessing

from featurebox.symbol.calculation.coefficient import try_add_coef
from featurebox.symbol.calculation.translate import general_expr
from featurebox.symbol.functions.dimfunc import dless

from featurebox.symbol.base import CalculatePrecisionSet
from featurebox.symbol.base import SymbolSet
from featurebox.symbol.base import SymbolTree
from featurebox.symbol.gp import cxOnePoint, varAnd, genGrow, staticLimit, mutShrink, selKbestDim
from featurebox.tools.packbox import Toolbox


import time

from numpy import random
from sklearn.datasets import load_boston

from featurebox.symbol.base import SymbolSet, SymbolTree, CalculatePrecisionSet
from featurebox.symbol.functions.dimfunc import dless, Dim
from featurebox.symbol.functions.npfunc import np_map
from featurebox.symbol.calculation.scores import compile_context, calculate_y


if __name__ == "__main__":
    # data
    data = load_boston()
    x = data["data"]
    y = data["target"]
    c = [6, 3, 4]
    # unit
    from sympy.physics.units import kg

    x_u = [kg] * 13
    y_u = kg
    c_u = [dless, dless, dless]

    x, x_dim = Dim.convert_x(x, x_u, target_units=None, unit_system="SI")
    y, y_dim = Dim.convert_xi(y, y_u)
    c, c_dim = Dim.convert_x(c, c_u)

    sc = preprocessing.MinMaxScaler((1,2))
    x = sc.fit_transform(x)
    t  = time.time()

    # symbolset
    pset0 = SymbolSet()
    pset0.add_features(x, y, x_dim=x_dim, y_dim=y_dim, x_group=2,
                       feature_name=["Ss%i" % i for i in range(13)])
    pset0.add_constants(c, c_dim=c_dim, c_prob=None)
    pset0.add_operations(power_categories=(2,3,4),
                         categories=("Add", "Mul", "Sub", "Div", "ln","exp"),
                         self_categories=None)


    pset0 = CalculatePrecisionSet(pset0, scoring=None, score_pen=(1,), filter_warning=True, cal_dim=False,
                                  dim_type=None,
                                  fuzzy=False, add_coef=False, inter_add=True,
                                  inner_add=False, n_jobs=1, batch_size=20,
                                  tq=True)
    sl = [SymbolTree.genGrow(pset0, 3, 4) for _ in range(50)]
    sll =[compile_context(o, pset0.context, pset0.gro_ter_con, simplify=True) for o  in sl]

    slll =[try_add_coef(i, pset0.data_x, pset0.y, pset0.terminals_and_constants_repr,
                 filter_warning=True, inter_add=True, inner_add=False, vector_add=True, np_maps=np_map())[1]
                for i in sll]
    sllll = [general_expr(o, pset0, simplifying=False) for o in slll]
    # tree="MSub(exp(gx5))"
    # sll = compile_context(tree, pset0.context, pset0.gro_ter_con, simplify=True)
    # slll = try_add_coef(sll, pset0.data_x, pset0.y, pset0.terminals_and_constants_repr,
    #               filter_warning=True, inter_add=False, inner_add=False, vector_add=True, np_maps=np_map())[1]
    # sllll=general_expr(slll, pset0, simplifying=False)



#!/usr/bin/python
# coding:utf-8

# @author: wangchangxin
# @contact: 986798607@qq.com
# @software: PyCharm
# @file: scores.py
# @License: GNU Lesser General Public License v3.0
"""
Notes:
    score method.
"""

import copy
import warnings

import numpy as np
import sympy
from sklearn.exceptions import DataConversionWarning
from sklearn.metrics import r2_score
from sklearn.model_selection import KFold
from sklearn.utils import check_array

from featurebox.symbol.calculation.coefficient import try_add_coef
from featurebox.symbol.calculation.translate import compile_context
from featurebox.symbol.functions.dimfunc import dim_map, dless, dnan, Dim


def calculate_y(expr01, x, y, terminals, add_coef=True, x_test=None, y_test=None,
                filter_warning=True, inter_add=True, inner_add=False, vector_add=False, np_maps=None):
    if filter_warning:
        warnings.filterwarnings("ignore")
    try:
        if add_coef:
            pre_y, expr01 = try_add_coef(expr01, x, y, terminals,
                                         filter_warning=filter_warning,
                                         inter_add=inter_add, inner_add=inner_add,
                                         vector_add=vector_add, np_maps=np_maps)
        else:
            func0 = sympy.utilities.lambdify(terminals, expr01, modules=[np_maps, "numpy"])
            pre_y = func0(*x)

        if x_test is not None and y_test is not None:
            func0 = sympy.utilities.lambdify(terminals, expr01, modules=[np_maps, "numpy"])
            pre_y = func0(*x_test)
            pre_y = pre_y.ravel()
            assert y_test.shape == pre_y.shape
            pre_y = check_array(pre_y, ensure_2d=False)
        else:
            pre_y = pre_y.ravel()
            assert y.shape == pre_y.shape
            pre_y = check_array(pre_y, ensure_2d=False)

    except (DataConversionWarning, AssertionError, ValueError, AttributeError, KeyError, ZeroDivisionError):
        pre_y = None

    return pre_y, expr01


def calculate_y_unpack(expr01, x, terminals):
    try:
        func0 = sympy.utilities.lambdify(terminals, expr01)
        pre_y = func0(*x)
        pre_y = pre_y.ravel()
        pre_y = check_array(pre_y, ensure_2d=False)

    except (DataConversionWarning, AssertionError, ValueError, AttributeError, KeyError, ZeroDivisionError):
        pre_y = None
    return pre_y


def uniform_score(score_pen=1):
    """return the worse score"""
    if score_pen >= 0:
        return -np.inf
    elif score_pen <= 0:
        return np.inf
    elif score_pen == 0:
        return 0
    else:
        return score_pen


def calculate_score(expr01, x, y, terminals, scoring=None, score_pen=(1,),
                    add_coef=True, filter_warning=True, inter_add=True,
                    inner_add=False, vector_add=False, np_maps=None):
    """

    Parameters
    ----------
    vector_add
    expr01: Expr
        sympy expression.
    x: list of np.ndarray
        list of xi
    y: np.ndarray
        y value
    terminals: list of sympy.Symbol
        features and constants
    scoring: list of Callbale, default is [sklearn.metrics.r2_score,]
        See Also sklearn.metrics
    score_pen: tuple of  1 or -1
        1 : best is positive, worse -np.inf
        -1 : best is negative, worse np.inf
        0 : best is positive , worse 0
    add_coef: bool
        bool
    filter_warning: bool
        bool
    inter_add: bool
        bool
    inner_add: bool
        bool
    np_maps: Callable
        user np.ndarray function

    Returns
    -------
    score:float
        score
    expr01: Expr
        New expr.
    pre_y: np.ndarray or float
        np.array or None
    """
    if filter_warning:
        warnings.filterwarnings("ignore")
    if not scoring:
        scoring = [r2_score, ]
    if not isinstance(scoring, (tuple, list)):
        scoring = [scoring, ]
    if isinstance(score_pen, int):
        score_pen = [score_pen, ]

    assert len(scoring) == len(score_pen), "the scoring and score_pen with same size"

    pre_y, expr01 = calculate_y(expr01, x, y, terminals, add_coef=add_coef,
                                filter_warning=filter_warning, inter_add=inter_add, inner_add=inner_add,
                                vector_add=vector_add,
                                np_maps=np_maps)

    try:
        sc_all = []
        for si, sp in zip(scoring, score_pen):
            sc = si(y, pre_y)
            if np.isnan(sc):
                sc = uniform_score(score_pen=sp)
            sc_all.append(sc)

    except (ValueError, RuntimeWarning):

        sc_all = [uniform_score(score_pen=i) for i in score_pen]

    return sc_all, expr01, pre_y


def calculate_cv_score(expr01, x, y, terminals, scoring=None, score_pen=(1,), cv=5, refit=True,
                       add_coef=True, filter_warning=True, inter_add=True,
                       inner_add=False, vector_add=False, np_maps=None):
    """
    use cv spilt for score,return the mean_test_score.
    use cv spilt for predict,return the cv_predict_y.(not be used)
    Notes:
        if cv and refit, all the data is refit to determination the coefficients.
        Thus the expression is not compact with the this scores, when re-calculated by this expression
    Parameters
    ----------
    refit: True:
        use forced, refit the coefficient use all data.
    cv:sklearn.model_selection._split._BaseKFold,int
        the shuffler must be False
    vector_add
    expr01: Expr
        sympy expression.
    x: list of np.ndarray
        list of xi
    y: np.ndarray
        y value
    terminals: list of sympy.Symbol
        features and constants
    scoring: list of Callbale, default is [sklearn.metrics.r2_score,]
        See Also sklearn.metrics
    score_pen: tuple of  1 or -1
        1 : best is positive, worse -np.inf
        -1 : best is negative, worse np.inf
        0 : best is positive , worse 0
    add_coef: bool
        bool
    filter_warning: bool
        bool
    inter_add: bool
        bool
    inner_add: bool
        bool
    np_maps: Callable
        user np.ndarray function

    Returns
    -------
    score:float
        score
    expr01: Expr
        New expr.
    pre_y: np.ndarray or float
        np.array or None
    """
    if filter_warning:
        warnings.filterwarnings("ignore")

    if cv == 1:
        sc_all, expr01, pre_y = calculate_score(expr01, x, y, terminals, scoring=scoring, score_pen=score_pen,
                                                add_coef=add_coef, filter_warning=filter_warning, inter_add=inter_add,
                                                inner_add=inner_add, vector_add=vector_add, np_maps=np_maps)
        return sc_all, expr01, pre_y

    if isinstance(cv, int):
        cv = KFold(cv, shuffle=False)

    cv_sc_all = []
    cv_expr01 = []
    cv_pre_y = []
    xx = [xi for xi in x if isinstance(xi, np.ndarray)]
    c = [xi for xi in x if not isinstance(xi, np.ndarray)]
    xx = [xi.reshape((-1, 1)) if xi.ndim == 1 else xi.T for xi in xx]

    for train_index, test_index in cv.split(xx[0], y):

        X_train = [xi[train_index] for xi in xx]
        X_test = [xi[test_index] for xi in xx]
        y_train, y_test = y[train_index], y[test_index]

        X_train.reverse()
        X_test.reverse()
        nc = copy.deepcopy(c)
        nc.reverse()
        X_train = [X_train.pop() if isinstance(xi, np.ndarray) else nc.pop() for index, xi in enumerate(x)]
        nc = copy.deepcopy(c)
        nc.reverse()
        X_test = [X_test.pop() if isinstance(xi, np.ndarray) else nc.pop() for index, xi in enumerate(x)]

        pre_y, expr01 = calculate_y(expr01, X_train, y_train, terminals,
                                    x_test=X_test, y_test=y_test,
                                    add_coef=add_coef,
                                    filter_warning=filter_warning, inter_add=inter_add,
                                    inner_add=inner_add,
                                    vector_add=vector_add,
                                    np_maps=np_maps)

        try:
            sc_all = []
            for si, sp in zip(scoring, score_pen):
                sc = si(y_test, pre_y)
                if np.isnan(sc):
                    sc = uniform_score(score_pen=sp)
                sc_all.append(sc)

        except (ValueError, RuntimeWarning):

            sc_all = [uniform_score(score_pen=i) for i in score_pen]

        cv_sc_all.append(sc_all)
        cv_expr01.append(expr01)
        cv_pre_y.append(pre_y)

    sc_all = list(np.mean(np.array(cv_sc_all), axis=0))
    try:
        pre_y = np.concatenate(cv_pre_y)
    except ValueError:
        pre_y = None

    if refit is True:
        "the refit only use for see the detial of calcualtion after loop"
        sc_all0, expr01, pre_y0 = calculate_score(expr01, x, y, terminals, scoring=scoring, score_pen=score_pen,
                                                  add_coef=add_coef, filter_warning=filter_warning, inter_add=inter_add,
                                                  inner_add=inner_add, vector_add=vector_add, np_maps=np_maps)

    return sc_all, expr01, pre_y


def score_dim(dim_, dim_type, fuzzy=False):
    if dim_type is None:
        return 1
    elif isinstance(dim_type, str):
        if dim_type == 'integer':
            return 1 if dim_.isinteger() else 0
        elif dim_type == 'coef':
            return 1 if not dim_.anyisnan() else 0
        else:
            raise TypeError("dim_type should be None,'coef', 'integer', special Dim or list of Dim")
    elif isinstance(dim_type, list):
        return 1 if dim_ in dim_type else 0
    elif isinstance(dim_type, Dim):
        if fuzzy:
            return 1 if dim_.is_same_base(dim_type) else 0
        else:
            return 1 if dim_ == dim_type else 0

    else:
        print(dim_type)
        raise TypeError("dim_type should be None,'coef','integer', special Dim or list of Dim")


def calcualte_dim(expr01, terminals, dim_list, dim_maps=None):
    """

    Parameters
    ----------
    expr01: Expr
        sympy expression.
    terminals: list of sympy.Symbol
        features and constants
    dim_list: list of Dim
        dims of features and constants
    dim_maps: Callable
        user dim_maps

    Returns
    -------
    Dim:
        dimension
    dim_score
        is target dim or not
    """
    terminals = [str(i) for i in terminals]
    if not dim_maps:
        dim_maps = dim_map()
    func0 = sympy.utilities.lambdify(terminals, expr01, modules=[dim_maps])
    try:
        dim_ = func0(*dim_list)
    except (ValueError, TypeError, ZeroDivisionError):
        dim_ = dnan
    if isinstance(dim_, (float, int)):
        dim_ = dless
    if not isinstance(dim_, Dim):
        dim_ = dnan

    return dim_


def calcualte_dim_score(expr01, terminals, dim_list, dim_type, fuzzy, dim_maps=None):
    """

    Parameters
    ----------
    expr01: Expr
        sympy expression.
    terminals: list of sympy.Symbol
        features and constants
    dim_list: list of Dim
        dims of features and constants
    dim_maps: Callable
        user dim_maps
    dim_type:list of Dim
        target dim
    fuzzy:
        fuzzy dim or not

    Returns
    -------
    Dim:
        dimension
    dim_score
        is target dim or not
    """
    dim_ = calcualte_dim(expr01, terminals, dim_list, dim_maps=dim_maps)

    dim_score = score_dim(dim_, dim_type, fuzzy)
    return dim_, dim_score


def calculate_collect(ind, context, x, y, terminals_and_constants_repr, gro_ter_con,
                      dim_ter_con_list, dim_type, fuzzy, cv=1, refit=True,
                      scoring=None, score_pen=(1,),
                      add_coef=True, filter_warning=True, inter_add=True, inner_add=False,
                      vector_add=False,
                      np_maps=None, dim_maps=None, cal_dim=True):
    expr01 = compile_context(ind, context, gro_ter_con)

    score, expr01, pre_y = calculate_cv_score(expr01, x, y, terminals_and_constants_repr,
                                              cv=cv, refit=refit,
                                              add_coef=add_coef, inter_add=inter_add,
                                              inner_add=inner_add, vector_add=vector_add,
                                              scoring=scoring, score_pen=score_pen,
                                              filter_warning=filter_warning,
                                              np_maps=np_maps)

    if cal_dim:
        dim, dim_score = calcualte_dim_score(expr01, terminals_and_constants_repr,
                                             dim_ter_con_list, dim_type, fuzzy,
                                             dim_maps=dim_maps)
    else:
        dim, dim_score = dless, 1

    return score, dim, dim_score



#!/usr/bin/python3.7
# -*- coding: utf-8 -*-

# @Time   : 2019/7/29 19:44
# @Author : Administrator
# @Software: PyCharm
# @License: BSD 3-Clause

"""
some tools for characterization
"""
import inspect
import numbers
import random
import time
from collections.abc import Iterable
from functools import partial, wraps

import numpy as np
from joblib import Parallel, delayed, effective_n_jobs
from tqdm import tqdm


def time_this_function(func):
    """
    time the function. use as a decorator.

    Parameters
    ----------
    func: Callable
        function

    Returns
    -------
    result
        function results
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(func.__name__, "time", end - start)
        return result

    return wrapper


def check_random_state(seed):
    """
    Turn seed into a random.RandomState instance

    Parameters
    ----------
    seed: None,int or RandomState
        If seed is None, return the RandomState singleton used by random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.

    Returns
    -------
    random.Random
        RandomState object
    """

    if seed is None or seed is random.random:
        return random.Random()
    if isinstance(seed, (numbers.Integral, np.integer)):
        return random.Random(seed)
    if isinstance(seed, random.Random):
        return seed
    raise ValueError('%r cannot be used to seed a seed'
                     ' instance' % seed)


def parallelize(n_jobs, func, iterable, respective=False, tq=True, batch_size='auto', **kwargs):
    """
    parallelize the function for iterable.

    make sure in if __name__ == "__main__":

    Parameters
    ----------
    batch_size
    respective:bool
        Import the parameters respectively or as a whole
    tq:bool
         View Progress or not
    n_jobs:int
        cpu numbers. n_jobs is the number of workers requested by the callers. Passing n_jobs=-1
    means requesting all available workers for instance matching the number of CPU cores on the worker host(s).
    func:
        function to calculate
    iterable:
        interable object
    kwargs:
        kwargs for function

    Returns
    -------
    results
        function results

    """

    func = partial(func, **kwargs)
    if effective_n_jobs(n_jobs) == 1:
        parallel, func = list, func
    else:
        parallel = Parallel(n_jobs=n_jobs, batch_size=batch_size)
        func = delayed(func)
    if tq:
        if respective:
            return parallel(func(*iter_i) for iter_i in tqdm(iterable))
        else:
            return parallel(func(iter_i) for iter_i in tqdm(iterable))
    else:
        if respective:
            return parallel(func(*iter_i) for iter_i in iterable)
        else:
            return parallel(func(iter_i) for iter_i in iterable)


def logg(func, printing=True, back=False):
    """
    Get the name of function

    Parameters
    ----------
    func:Callable
        function to calculate
    printing:bool
        print or not
    back:bool
        return result or not

    Returns
    -------
    function results
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if inspect.isclass(func):
            name = "instance of %s" % func.__name__
            arg_dict = ""
            if printing:
                print(name, arg_dict)
            result = func(*args, **kwargs)
        elif inspect.isfunction(func):
            arg_dict = inspect.getcallargs(func, *args, **kwargs)
            name = func.__name__
            if printing:
                print(name, arg_dict)
            result = func(*args, **kwargs)
        else:
            arg_dict = ""
            name = ""
            result = func(*args, **kwargs)
            pass
        if back:
            return (name, arg_dict), result
        else:
            return result

    return wrapper


def name_to_name(*iters, search, search_which=1, return_which=(1,), two_layer=False):
    """

    Parameters
    ----------
    iters:tuple
        iterable objects to select and sort
    search:Iterable
        the rank basis
    search_which:int
        the (the index of iters) of the rank basis. where to rank
    return_which:tuple of int
        return index in inters
    two_layer:bool
        search is nested with two layer or not

    Returns
    -------
    result:Iterable
        Result of find and sort
    """
    if isinstance(return_which, int):
        return_which = tuple([return_which, ])
    if two_layer:

        results_all = []
        if isinstance(search, Iterable):
            for index_i in search:
                results_all.append(
                    name_to_name(*iters, search=index_i, search_which=search_which,
                                 return_which=return_which, two_layer=False))

            if len(return_which) >= 2:
                return list(zip(*results_all))
            else:
                return results_all
        else:
            raise IndexError("search_name or search should be iterable")

    else:

        zeros = [list(range(len(iters[0])))]

        zeros.extend([list(_) for _ in iters])

        iters = zeros

        zips = list(zip(*iters))

        if isinstance(search, Iterable):

            search_index = [iters[search_which].index(i) for i in search]

            results = [zips[i] for i in search_index]

        else:

            raise IndexError("search_name or search should be iterable")

        res = list(zip(*results))

        if not res:
            return_res = [[] for _ in return_which]
        else:
            return_res = [res[_] for _ in return_which]

        if len(return_which) == 1:
            return_res = return_res[0]
        return return_res


class _TTClass(dict):
    def __init__(self, **kwargs):
        super(_TTClass, self).__init__(**kwargs)

    def _t(self):
        self._r()

    def _p(self):
        a = np.array(list(self.keys()))
        b = np.array(list(self.values()))
        a0 = np.delete(a, 0)
        a = np.delete(a, -1)
        b0 = np.delete(b, 0)
        b = np.delete(b, -1)
        ti = b0 - b
        func = lambda x, y: "{}-{}:".format(x, y)
        ufunc = np.frompyfunc(func, 2, 1)
        ni = ufunc(a0, a)
        re = np.vstack((ni, ti))
        re = re.T
        print(re)
        self.clear()

    def _r(self, name=None):
        ti = time.time()
        if name is None:
            n = len(self)
            name = "t%s" % n
        self[name] = ti

    def subs(self, name1, name2):
        if name2 in self and name2 in self:
            return self[name1] - self[name2]
        else:
            raise NameError("There is no name:{} or {}".format(name1, name2))

    # def tp(self, name=None):
    #     ti = time.time()
    #     n = len(self)
    #     if name is None:
    #         name = "tp"
    #     if n >= 1:
    #         name0 = list(self.keys())[-1]
    #         t0 = list(self.values())[-1]
    #         print("-".join((name,name0)), ti - t0)
    #     else:
    #         raise UserWarning("The .tp only used after .t, or .r")


class TTClass(_TTClass):
    """
    quick time.
    use tt object.
    >>>tt.t
    >>>#tt.**n
    >>>tt.p
    """

    def __init__(self, **kwargs):
        super(_TTClass, self).__init__(**kwargs)

    def __getattribute__(self, item):
        if item is "t":
            _TTClass._t(self)

        elif item is "p":
            _TTClass._p(self)

        elif item[-1] in "0123456789":
            _TTClass._r(self, name=item)

        else:
            return _TTClass.__getattribute__(self, item)


tt = TTClass()
